OAUTH2_V2_DISCVOCERY_DOC = {'auth': {'oauth2': {'scopes': {'https://www.googleapis.com/auth/plus.login': {'description': 'Know the list of people in your circles, your age range, and '
                                                                                              'language'},
                                'https://www.googleapis.com/auth/plus.me': {'description': 'Know who you are on Google'},
                                'https://www.googleapis.com/auth/userinfo.email': {'description': 'View your email address'},
                                'https://www.googleapis.com/auth/userinfo.profile': {'description': 'View your basic profile info'}}}},
 'basePath': '/',
 'baseUrl': 'https://www.googleapis.com/',
 'batchPath': 'batch/oauth2/v2',
 'description': 'Obtains end-user authorization grants for use with other Google APIs.',
 'discoveryVersion': 'v1',
 'documentationLink': 'https://developers.google.com/accounts/docs/OAuth2',
 'etag': '"J3WqvAcMk4eQjJXvfSI4Yr8VouA/fwW2utFkCnVanfAd4kn3jkF7-Z8"',
 'icons': {'x16': 'https://www.gstatic.com/images/branding/product/1x/googleg_16dp.png',
           'x32': 'https://www.gstatic.com/images/branding/product/1x/googleg_32dp.png'},
 'id': 'oauth2:v2',
 'kind': 'discovery#restDescription',
 'methods': {'getCertForOpenIdConnect': {'httpMethod': 'GET', 'id': 'oauth2.getCertForOpenIdConnect', 'path': 'oauth2/v2/certs', 'response': {'$ref': 'Jwk'}},
             'tokeninfo': {'httpMethod': 'POST',
                           'id': 'oauth2.tokeninfo',
                           'parameters': {'access_token': {'location': 'query', 'type': 'string'},
                                          'id_token': {'location': 'query', 'type': 'string'},
                                          'token_handle': {'location': 'query', 'type': 'string'}},
                           'path': 'oauth2/v2/tokeninfo',
                           'response': {'$ref': 'Tokeninfo'}}},
 'name': 'oauth2',
 'ownerDomain': 'google.com',
 'ownerName': 'Google',
 'parameters': {'alt': {'default': 'json',
                        'description': 'Data format for the response.',
                        'enum': ['json'],
                        'enumDescriptions': ['Responses with Content-Type of application/json'],
                        'location': 'query',
                        'type': 'string'},
                'fields': {'description': 'Selector specifying which fields to include in a partial response.', 'location': 'query', 'type': 'string'},
                'key': {'description': 'API key. Your API key identifies your project and provides you with API access, quota, and reports. Required unless '
                                       'you provide an OAuth 2.0 token.',
                        'location': 'query',
                        'type': 'string'},
                'oauth_token': {'description': 'OAuth 2.0 token for the current user.', 'location': 'query', 'type': 'string'},
                'prettyPrint': {'default': 'true',
                                'description': 'Returns response with indentations and line breaks.',
                                'location': 'query',
                                'type': 'boolean'},
                'quotaUser': {'description': 'An opaque string that represents a user for quota purposes. Must not exceed 40 characters.',
                              'location': 'query',
                              'type': 'string'},
                'userIp': {'description': 'Deprecated. Please use quotaUser instead.', 'location': 'query', 'type': 'string'}},
 'protocol': 'rest',
 'resources': {'userinfo': {'methods': {'get': {'httpMethod': 'GET',
                                                'id': 'oauth2.userinfo.get',
                                                'path': 'oauth2/v2/userinfo',
                                                'response': {'$ref': 'Userinfoplus'},
                                                'scopes': ['https://www.googleapis.com/auth/plus.login',
                                                           'https://www.googleapis.com/auth/plus.me',
                                                           'https://www.googleapis.com/auth/userinfo.email',
                                                           'https://www.googleapis.com/auth/userinfo.profile']}},
                            'resources': {'v2': {'resources': {'me': {'methods': {'get': {'httpMethod': 'GET',
                                                                                          'id': 'oauth2.userinfo.v2.me.get',
                                                                                          'path': 'userinfo/v2/me',
                                                                                          'response': {'$ref': 'Userinfoplus'},
                                                                                          'scopes': ['https://www.googleapis.com/auth/plus.login',
                                                                                                     'https://www.googleapis.com/auth/plus.me',
                                                                                                     'https://www.googleapis.com/auth/userinfo.email',
                                                                                                     'https://www.googleapis.com/auth/userinfo.profile']}}}}}}}},
 'revision': '20181105',
 'rootUrl': 'https://www.googleapis.com/',
 'schemas': {'Jwk': {'id': 'Jwk',
                     'properties': {'keys': {'items': {'properties': {'alg': {'default': 'RS256', 'type': 'string'},
                                                                      'e': {'type': 'string'},
                                                                      'kid': {'type': 'string'},
                                                                      'kty': {'default': 'RSA', 'type': 'string'},
                                                                      'n': {'type': 'string'},
                                                                      'use': {'default': 'sig', 'type': 'string'}},
                                                       'type': 'object'},
                                             'type': 'array'}},
                     'type': 'object'},
             'Tokeninfo': {'id': 'Tokeninfo',
                           'properties': {'access_type': {'description': 'The access type granted with this token. It can be offline or online.',
                                                          'type': 'string'},
                                          'audience': {'description': 'Who is the intended audience for this token. In general the same as issued_to.',
                                                       'type': 'string'},
                                          'email': {'description': 'The email address of the user. Present only if the email scope is present in the request.',
                                                    'type': 'string'},
                                          'expires_in': {'description': 'The expiry time of the token, as number of seconds left until expiry.',
                                                         'format': 'int32',
                                                         'type': 'integer'},
                                          'issued_to': {'description': 'To whom was the token issued to. In general the same as audience.', 'type': 'string'},
                                          'scope': {'description': 'The space separated list of scopes granted to this token.', 'type': 'string'},
                                          'token_handle': {'description': 'The token handle associated with this token.', 'type': 'string'},
                                          'user_id': {'description': 'The obfuscated user id.', 'type': 'string'},
                                          'verified_email': {'description': 'Boolean flag which is true if the email address is verified. Present only if the '
                                                                            'email scope is present in the request.',
                                                             'type': 'boolean'}},
                           'type': 'object'},
             'Userinfoplus': {'id': 'Userinfoplus',
                              'properties': {'email': {'description': "The user's email address.", 'type': 'string'},
                                             'family_name': {'description': "The user's last name.", 'type': 'string'},
                                             'gender': {'description': "The user's gender.", 'type': 'string'},
                                             'given_name': {'description': "The user's first name.", 'type': 'string'},
                                             'hd': {'description': 'The hosted domain e.g. example.com if the user is Google apps user.', 'type': 'string'},
                                             'id': {'description': 'The obfuscated ID of the user.', 'type': 'string'},
                                             'link': {'description': 'URL of the profile page.', 'type': 'string'},
                                             'locale': {'description': "The user's preferred locale.", 'type': 'string'},
                                             'name': {'description': "The user's full name.", 'type': 'string'},
                                             'picture': {'description': "URL of the user's picture image.", 'type': 'string'},
                                             'verified_email': {'default': 'true',
                                                                'description': 'Boolean flag which is true if the email address is verified. Always verified '
                                                                               "because we only return the user's primary email address.",
                                                                'type': 'boolean'}},
                              'type': 'object'}},
 'servicePath': '',
 'title': 'Google OAuth2 API',
 'version': 'v2'}

WELLKNOWN_OPENID_CONFIGS = {'authorization_endpoint': 'https://accounts.google.com/o/oauth2/v2/auth',
 'claims_supported': ['aud',
                      'email',
                      'email_verified',
                      'exp',
                      'family_name',
                      'given_name',
                      'iat',
                      'iss',
                      'locale',
                      'name',
                      'picture',
                      'sub'],
 'code_challenge_methods_supported': ['plain', 'S256'],
 'id_token_signing_alg_values_supported': ['RS256'],
 'issuer': 'https://accounts.google.com',
 'jwks_uri': 'https://www.googleapis.com/oauth2/v3/certs',
 'response_types_supported': ['code', 'token', 'id_token', 'code token', 'code id_token', 'token id_token', 'code token id_token', 'none'],
 'revocation_endpoint': 'https://oauth2.googleapis.com/revoke',
 'scopes_supported': ['openid', 'email', 'profile'],
 'subject_types_supported': ['public'],
 'token_endpoint': 'https://oauth2.googleapis.com/token',
 'token_endpoint_auth_methods_supported': ['client_secret_post', 'client_secret_basic'],
 'userinfo_endpoint': 'https://openidconnect.googleapis.com/v1/userinfo'}


import json
import os

import pytest

from aiogoogle.resource import Resource, GoogleAPI, Method
from aiogoogle.models import Request
from aiogoogle.excs import ValidationError


from ..test_globals import ALL_APIS


# TODO: Test additionalParameters (Tricky)

def test_aiogoogle_compiles_discovery_re_pattern(create_api):
    pass

def test_validates_url_path_params(create_api):
    calendar = create_api('calendar', 'v3')
    assert calendar.acl.insert(calendarId='string', validate=True)

    with pytest.raises(ValidationError):
        calendar.acl.insert(calendarId=1, validate=True)
    with pytest.raises(ValidationError):
        calendar.acl.insert(calendarId=None, validate=True)
    with pytest.raises(ValidationError):
        calendar.acl.insert(calendarId=True, validate=True)
    with pytest.raises(ValidationError):
        calendar.acl.insert(calendarId=['asdasd'], validate=True)


def test_validates_url_query_params(create_api):
    youtube = create_api('youtube', 'v3')
    youtube.videos.list(part='snippet', validate=True)

    with pytest.raises(ValidationError):
        youtube.videos.list(part=0.8, validate=True)
    with pytest.raises(ValidationError):
        youtube.videos.list(part=False, validate=True)
    with pytest.raises(ValidationError):
        youtube.videos.list(part=1, validate=True)
    with pytest.raises(ValidationError):
        youtube.videos.list(part=['snippet'], validate=True)

def test_validates_body_data(create_api):
    youtube = create_api('youtube', 'v3')
    youtube.videos.insert(
        autoLevels=True,
        part='snippet',
        data=dict(ageGating=dict(alcoholContent=True)),
        validate=True
    )

    with pytest.raises(ValidationError):
        youtube.videos.insert(
            autoLevels=True,
            part='snippet',
            data=1.82,
            validate=True
        )

def test_validates_body_json(create_api):
    youtube = create_api('youtube', 'v3')
    youtube.videos.insert(
        autoLevels=True,
        part='snippet',
        json=dict(ageGating=dict(alcoholContent=True)),
        validate=True
    )

def test_validates_body_json_2(create_api):
    youtube = create_api('youtube', 'v3')
    with pytest.raises(ValidationError):
        youtube.videos.insert(
            autoLevels=True,
            part='snippet',
            json=False,
            validate=True
        )

def test_validates_body_json_3(create_api):
    youtube = create_api('youtube', 'v3')
    with pytest.raises(ValidationError):
        youtube.videos.insert(
            autoLevels=True,
            part='snippet',
            json=dict(ageGating=dict(alcoholContent='asdasdasdasd')),
            validate=True
        )

def test_validates_body_json_4(create_api):
    youtube = create_api('youtube', 'v3')
    with pytest.raises(ValidationError):
        youtube.videos.insert(
            autoLevels=True,
            part='snippet',
            json=dict(ageGating=dict(alcoholContent={'asdasdkj': 'asdads'})),
            validate=True
        )

def test_validates_body_json_5(create_api):
    youtube = create_api('youtube', 'v3')
    youtube.videos.insert(
        autoLevels=True,
        part='snippet',
        json=None,
        validate=True
    )

def test_validates_body_json_6(create_api):
    calendar = create_api('calendar', 'v3')
    calendar.calendarList.get(
        calendarId='asdasd',
        validate=True
    )

def test_validates_body_json_7(create_api):
    calendar = create_api('calendar', 'v3')
    with pytest.raises(ValidationError):
        calendar.calendarList.get(
            calendarId=True,
            validate=True
        )

def test_validates_body_json_8(create_api):
    calendar = create_api('calendar', 'v3')
    with pytest.raises(ValidationError):
        calendar.calendarList.get(
            calendarId=123,
            validate=True
        )

def test_validates_body_json_9(create_api):
    calendar = create_api('calendar', 'v3')
    calendar.calendarList.list(validate=True)
    calendar.calendarList.list._method_specs['request'] = calendar.calendarList.list._method_specs['response']  # Testing arrays and objects
    with pytest.raises(ValidationError):
        calendar.calendarList.list(
            json='asdasd',
            validate=True
        )

def test_validates_body_json_10(create_api):
    calendar = create_api('calendar', 'v3')
    calendar.calendarList.list(validate=True)
    calendar.calendarList.list._method_specs['request'] = calendar.calendarList.list._method_specs['response']  # Testing arrays and objects
    with pytest.raises(ValidationError):
        calendar.calendarList.list(
            json=[],
            validate=True
        )
def test_validates_body_json_11(create_api):
    calendar = create_api('calendar', 'v3')
    calendar.calendarList.list(validate=True)
    calendar.calendarList.list._method_specs['request'] = calendar.calendarList.list._method_specs['response']  # Testing arrays and objects
    calendar.calendarList.list(
        json={},
        validate=True
    )

def test_validates_body_json_12(create_api):
    calendar = create_api('calendar', 'v3')
    calendar.calendarList.list(validate=True)
    calendar.calendarList.list._method_specs['request'] = calendar.calendarList.list._method_specs['response']  # Testing arrays and objects
    with pytest.raises(ValidationError):
        calendar.calendarList.list(
            json={
                'items':{}
            }
        )

def test_validates_body_json_13(create_api):
    calendar = create_api('calendar', 'v3')
    calendar.calendarList.list(validate=True)
    calendar.calendarList.list._method_specs['request'] = calendar.calendarList.list._method_specs['response']  # Testing arrays and objects
    calendar.calendarList.list(
        json={
            'items':[
                {
                    'accessRole': 'a_valid_access_role'
                },
            ]
        }
    )

def test_validates_body_json_14(create_api):
    calendar = create_api('calendar', 'v3')
    calendar.calendarList.list(validate=True)
    calendar.calendarList.list._method_specs['request'] = calendar.calendarList.list._method_specs['response']  # Testing arrays and objects
    with pytest.raises(ValidationError):
        calendar.calendarList.list(
            json={
                'items':[
                    {
                        'accessRole': 1
                    },
                ]
            }
        )

def test_validates_body_json_15(create_api):
    calendar = create_api('calendar', 'v3')
    calendar.calendarList.list(validate=True)
    calendar.calendarList.list._method_specs['request'] = calendar.calendarList.list._method_specs['response']  # Testing arrays and objects
    with pytest.warns(UserWarning, match='this_isnt_supposed_to_be_here'):
        calendar.calendarList.list(
            json={
                'items':[
                    {
                        'accessRole': 'a_valid_access_role'
                    },
                    {
                        'this_isnt_supposed_to_be_here': True
                    }
                ]
            }
        )

def test_validates_body_json_16(create_api):
    calendar = create_api('calendar', 'v3')
    calendar.calendarList.list(validate=True)
    calendar.calendarList.list._method_specs['request'] = calendar.calendarList.list._method_specs['response']  # Testing arrays and objects
    with pytest.warns(UserWarning, match='this_isnt_supposed_to_be_here'):
        calendar.calendarList.list(
            json={
                'items':[
                    {
                        'accessRole': 'a_valid_access_role'
                    },
                    {
                        'this_isnt_supposed_to_be_here': 'asdasdasd'
                    }
                ]
            }
        )

def test_validates_body_json_17(create_api):
    calendar = create_api('calendar', 'v3')
    calendar.calendarList.list(validate=True)
    calendar.calendarList.list._method_specs['request'] = calendar.calendarList.list._method_specs['response']  # Testing arrays and objects
    with pytest.warns(UserWarning, match='this_isnt_supposed_to_be_here'):
        calendar.calendarList.list(
            json={
                'items':[
                    {
                        'this_isnt_supposed_to_be_here': True
                    }
                ]
            }
        )

def test_validates_body_json_18(create_api):
    calendar = create_api('calendar', 'v3')
    calendar.calendarList.list(validate=True)
    calendar.calendarList.list._method_specs['request'] = calendar.calendarList.list._method_specs['response']  # Testing arrays and objects
    with pytest.warns(UserWarning, match='this_isnt_supposed_to_be_here'):
        calendar.calendarList.list(
            json={
                'items':[
                    {
                        'this_isnt_supposed_to_be_here': 'asdasdasd'
                    }
                ]
            }
        )

def test_validates_body_json_19(create_api):
    calendar = create_api('calendar', 'v3')
    calendar.calendarList.list(validate=True)
    calendar.calendarList.list._method_specs['request'] = calendar.calendarList.list._method_specs['response']  # Testing arrays and objects
    with pytest.warns(UserWarning, match='this_isnt_supposed_to_be_here'):
        with pytest.raises(ValidationError):
            calendar.calendarList.list(
                json={
                    'items':[
                        {
                            'accessRole': 1,
                            'this_isnt_supposed_to_be_here': 123
                        },
                    ]
                }
            )

def test_validates_body_json_20(create_api):
    calendar = create_api('calendar', 'v3')
    calendar.calendarList.list(validate=True)
    calendar.calendarList.list._method_specs['request'] = calendar.calendarList.list._method_specs['response']  # Testing arrays and objects
    with pytest.warns(UserWarning, match='this_isnt_supposed_to_be_here'):
        with pytest.raises(ValidationError):
            calendar.calendarList.list(
                json={
                    'items':[
                        {
                            'accessRole': 1,
                            'this_isnt_supposed_to_be_here': 'asdsda'
                        },
                    ]
                }
            )

def test_validates_body_json_21(create_api):
    calendar = create_api('calendar', 'v3')
    calendar.calendarList.list(validate=True)
    calendar.calendarList.list._method_specs['request'] = calendar.calendarList.list._method_specs['response']  # Testing arrays and objects
    calendar.calendarList.list(
        json={
            'etag': 'asd',
            'items':[
                {
                    'accessRole': 'asdasd',
                    'defaultReminders': [
                        {
                            'method': 'asd',
                            'minutes': 1
                        },
                        {
                            'method': 'asd',
                            'minutes': 1
                        },
                    ]
                },
            ]
        }
    )

def test_validates_body_json_22(create_api):
    calendar = create_api('calendar', 'v3')
    calendar.calendarList.list(validate=True)
    calendar.calendarList.list._method_specs['request'] = calendar.calendarList.list._method_specs['response']  # Testing arrays and objects
    with pytest.raises(ValidationError):
        calendar.calendarList.list(
            json={
                'etag': 'asd',
                'items':[
                    {
                        'accessRole': 'asdasd',
                        'defaultReminders': [
                            {
                                'method': 'asd',
                                'minutes': '123'
                            },
                            {
                                'method': 'asd',
                                'minutes': 1
                            },
                        ]
                    },
                ]
            }
        )

def test_validates_body_json_23(create_api):
    calendar = create_api('calendar', 'v3')
    calendar.calendarList.list(validate=True)
    calendar.calendarList.list._method_specs['request'] = calendar.calendarList.list._method_specs['response']  # Testing arrays and objects
    with pytest.raises(ValidationError):
        calendar.calendarList.list(
            json={
                'etag': 'asd',
                'items':[
                    {
                        'accessRole': 'asdasd',
                        'defaultReminders': [
                            {
                                'method': 'asd',
                                'minutes': 1
                            },
                            {
                                'method': 'asd',
                                'minutes': 1
                            },
                            'asd'
                        ]
                    },
                ]
            }
        )

def test_validates_body_json_24(create_api):
    calendar = create_api('calendar', 'v3')
    calendar.calendarList.list(validate=True)
    calendar.calendarList.list._method_specs['request'] = calendar.calendarList.list._method_specs['response']  # Testing arrays and objects
    with pytest.raises(ValidationError):
        calendar.calendarList.list(
            json={
                'etag': 'asd',
                'items':[
                    {
                        'accessRole': 3,
                        'defaultReminders': [
                            {
                                'method': 'asd',
                                'minutes': 1
                            },
                            {
                                'method': 'asd',
                                'minutes': 1
                            },
                        ]
                    },
                ]
            }
        )


__name__ = 'aiogoogle'
__about__ = "An async-framework agnostic async library for Google's Discovery Service"
__url__ = 'https://github.com/omarryhan/aiogoogle'
__version_info__ = ('0', '0', '1b1')
__version__ = '.'.join(__version_info__)
__author__ = 'Omar Ryhan'
__author_email__ = 'omarryhan@gmail.com'
__maintainer__ = 'Omar Ryhan'
__license__ = 'GNU'
__copyright__ = '(c) 2018 by Omar Ryhan'


from aiogoogle.auth.managers import Oauth2Manager
from aiogoogle.models import Request
from aiogoogle.resource import GoogleAPI
from aiogoogle.client import Aiogoogle
from aiogoogle.sessions.aiohttp_session import AiohttpSession
import pytest


@pytest.fixture
def manager():
    return Oauth2Manager(AiohttpSession)

def test_oauth2_manager_has_valid_oauth2_api(manager):
    assert isinstance(manager.oauth2_api, GoogleAPI)

def test_authorize(manager):
    access_name = 'access'
    unauth_req = Request()
    user_creds = {
        'access_token': access_name
    }

    auth_req = manager.authorize(unauth_req, user_creds)
    assert auth_req.headers['Authorization'] == 'Bearer ' + access_name

    with pytest.raises(KeyError):
        auth_req = manager.authorize(unauth_req, {})

def test_is_ready(manager):
    # client creds must have: client secret, client id, scopes, redirect uri
    ready_creds = {
        'client_secret': 'asasd',
        'client_id': 'casas',
        'scopes': ['asd', 'asd'],
        'redirect_uri': 'asdasd'
    }

    unready_creds = [
        {
            'client_secret': 'asasd',
            'scopes': ['asd', 'asd'],
            'redirect_uri': 'asdasd'
        },
        {
            'client_secret': 'asasd',
            'client_id': 'casas',
            'scopes': ['asd', 'asd'],
        },
        {
            'client_id': 'casas',
            'scopes': ['asd', 'asd'],
            'redirect_uri': 'asdasd'
        },
        {
            'client_secret': 'asasd',
            'client_id': 'casas',
            'scopes': 'asd',
            'redirect_uri': 'asdasd'
        },
        {
            'client_secret': 'asasd',
            'client_id': 'casas',
            'scopes': {'asd': 'asd'},
            'redirect_uri': 'asdasd'
        },
    ]

    assert manager.is_ready(ready_creds) is True

    for unready_creds_ in unready_creds:
        assert manager.is_ready(unready_creds_) is False

def test_authorization_url():
    # TODO:
    pass

def test_build_user_creds_req():
    # TODO:
    pass

def test_build_user_creds_from_res():
    # TODO:
    pass

def test_get_expires_at():
    # TODO:
    pass

def test_is_expired():
    # TODO:
    pass

def test_build_refresh_request():
    # TODO:
    pass

def test_build_revoke_request():
    # TODO:
    pass

def test_user_authorized_for_method(open_discovery_document):
    discovery_document = open_discovery_document('youtube', 'v3')
    youtube = GoogleAPI(discovery_document=discovery_document)
    aiogoogle = Aiogoogle()
    assert aiogoogle.oauth2.authorized_for_method(
        method=youtube.videos.list,
        user_creds={
            'scopes': [
                "https://www.googleapis.com/auth/youtube",
                "https://www.googleapis.com/auth/youtube.force-ssl",
                "https://www.googleapis.com/auth/youtube.readonly",
                "https://www.googleapis.com/auth/youtubepartner"
            ]
        }
    ) is True

    assert aiogoogle.oauth2.authorized_for_method(
        method=youtube.videos.list,
        user_creds={
            'scopes': [
                #"https://www.googleapis.com/auth/youtube",
                "https://www.googleapis.com/auth/youtube.force-ssl",
                "https://www.googleapis.com/auth/youtube.readonly",
                "https://www.googleapis.com/auth/youtubepartner"
            ]
        }
    ) is False

    assert aiogoogle.oauth2.authorized_for_method(
        method=youtube.videos.list,
        user_creds={
            'scopes': [
                "https://www.googleapis.com/auth/youtube",
                "https://www.googleapis.com/auth/youtube.force-ssl",
                "https://www.googleapis.com/auth/youtube.readonly",
                "https://www.googleapis.com/auth/youtubepartner",
                "useless_scope"
            ]
        }
    ) is True

    assert aiogoogle.oauth2.authorized_for_method(
        method=youtube.videos.list,
        user_creds={
            'scopes': []
        }
    ) is False

    with pytest.raises(TypeError) as e:
        aiogoogle.oauth2.authorized_for_method(
            method=youtube.videos.list,
            user_creds={
                'scopes': None
            }
        )
        assert str(e) == 'Scopes should be an instance of list, set or tuple'



import os
import json

import pytest

from aiogoogle.models import Request
from aiogoogle import Aiogoogle
from aiogoogle.resource import GoogleAPI


@pytest.fixture('function')
def open_discovery_document():
    def wrapped(name, version):
        current_dir = os.getcwd()
        file_name = current_dir + '/tests/data/' + name + '_' + version + '_discovery_doc.json'
        with open(file_name, 'r') as discovery_doc:
            discovery_doc_dict = json.loads(discovery_doc.read())
        return discovery_doc_dict
    return wrapped

@pytest.fixture('function')
def create_api(open_discovery_document):
    def wrapped(name, version):
        disc_doc = open_discovery_document(name, version)
        return GoogleAPI(discovery_document=disc_doc)
    return wrapped


import pytest

from aiogoogle import Aiogoogle
from aiogoogle.resource import Resource, GoogleAPI, Method, STACK_QUERY_PARAMETER_DEFAULT_VALUE, STACK_QUERY_PARAMETERS
from ..test_globals import ALL_APIS

@pytest.mark.parametrize('name,version', ALL_APIS)
def test_resource_constructor(open_discovery_document, name, version):
    discovery_document = open_discovery_document(name, version)
    api = GoogleAPI(discovery_document=discovery_document)
    for resource_name, _ in discovery_document.get('resources').items():
        resource_object = getattr(api, resource_name)
        assert isinstance(resource_object, Resource)
        assert resource_name == resource_object.name
        assert resource_object._resource_specs == discovery_document['resources'][resource_name]
        assert resource_object._global_parameters == discovery_document.get('parameters')
        assert resource_object._schemas == discovery_document.get('schemas')
        ##
        assert api['schemas'] is resource_object._schemas
        assert api['parameters'] is resource_object._global_parameters

def test_resources_available_property():
    resource = Resource(
        name='irrelevant',
        resource_specs={
            'resources': {
                'first_resource': 1,
                'second_resource': 2
            },
            'methods': {
                'third_method': 3,
                'forth_method': 4
            }
        },
        global_parameters = None,
        schemas = None,
        batch_path = None,
        root_url = None,
        validate = False,
        service_path = None
    )
    assert 'first_resource' in resource.resources_available
    assert 'second_resource' in resource.resources_available

def test_methods_available_property():
    resource = Resource(
        name='irrelevant',
        resource_specs={
            'resources': {
                'first_resource': 1,
                'second_resource': 2
            },
            'methods': {
                'third_method': 3,
                'forth_method': 4
            }
        },
        global_parameters = None,
        schemas = None,
        batch_path = None,
        root_url = None,
        validate = False,
        service_path = None
    )
    assert 'third_method' in resource.methods_available
    assert 'forth_method' in resource.methods_available

def test_resource_len():
    resource = Resource(
        name='irrelevant',
        resource_specs={
            'resources': {
                'first_resource': 1,
                'second_resource': 2,
                'basdasd': 239
            },
            'methods': {
                'third_method': 3,
                'forth_method': 4
            }
        },
        global_parameters = None,
        schemas = None,
        root_url = None,
        batch_path = None,
        validate = False,
        service_path = None
    )
    assert len(resource) == 2

@pytest.mark.parametrize('name,version', ALL_APIS)
def test_resource_returns_nested_resource(open_discovery_document, name, version):
    discovery_document = open_discovery_document(name, version)
    api = GoogleAPI(discovery_document=discovery_document)
    for resource_name, _ in discovery_document.get('resources', {}).items():
        resource = getattr(api, resource_name)
        if resource.resources_available:
            # Assert that it returns resources not methods
            for nested_resource_name in resource.resources_available:
                nested_resource = getattr(resource, nested_resource_name)
                assert isinstance(nested_resource, Resource)

@pytest.mark.parametrize('name,version', ALL_APIS)
def test_resource_returns_available_methods(open_discovery_document, name, version):
    discovery_document = open_discovery_document(name, version)
    api = GoogleAPI(discovery_document=discovery_document)
    for resource_name, _ in discovery_document.get('resources', {}).items():
        resource = getattr(api, resource_name)
        if resource.methods_available:
            # Assert that it returns resources not methods
            for available_method_name in resource.methods_available:
                available_method = getattr(resource, available_method_name)
                assert isinstance(available_method, Method)


@pytest.mark.parametrize('name,version', ALL_APIS)
def test_method_construction(open_discovery_document, name, version):
    discovery_document = open_discovery_document(name, version)
    api = GoogleAPI(discovery_document=discovery_document)
    for resource_name, _ in discovery_document.get('resources').items():
        resource = getattr(api, resource_name)
        for method_name in resource.methods_available:
            method = getattr(resource, method_name)
            # Assert a resource has the returned method
            assert method.name in discovery_document['resources'][resource_name]['methods']
            # Assert specs belong to the created method
            assert method._method_specs == discovery_document['resources'][resource_name]['methods'][method_name]
            # Assert global parameters were passed correctly
            assert method._global_parameters == discovery_document.get('parameters')
            # Assert schemas where passed correctly
            assert method._schemas == discovery_document.get('schemas')

@pytest.mark.parametrize('name,version', ALL_APIS)
def test_resource_construction(open_discovery_document, name, version):
    discovery_document = open_discovery_document(name, version)
    api = GoogleAPI(discovery_document=discovery_document)
    for resource_name, _ in discovery_document.get('resources').items():
        resource = getattr(api, resource_name)
        for nested_resource_name in resource.resources_available:
            nested_resource = getattr(resource, nested_resource_name)
            assert isinstance(nested_resource, Resource)

@pytest.mark.parametrize('name,version', ALL_APIS)
def test_calling_resource_fails(open_discovery_document, name, version):
    discovery_document = open_discovery_document(name, version)
    api = GoogleAPI(discovery_document=discovery_document)
    for resource_name, _ in discovery_document.get('resources').items():
        resource = getattr(api, resource_name)
        with pytest.raises(TypeError):
            resource()

@pytest.mark.parametrize('name,version', ALL_APIS)
def test_str_resource(open_discovery_document, name, version):
    discovery_document = open_discovery_document(name, version)
    api = GoogleAPI(discovery_document=discovery_document)
    for resource_name, _ in discovery_document.get('resources').items():
        resource = getattr(api, resource_name)
        assert discovery_document['rootUrl'] in str(resource)
        assert 'resource' in str(resource)

@pytest.mark.parametrize('name,version', ALL_APIS)
def test_stack_parameters(open_discovery_document, name, version):
    discovery_document = open_discovery_document(name, version)
    api = GoogleAPI(discovery_document=discovery_document)
    for resource_name, _ in discovery_document.get('resources').items():
        resource = getattr(api, resource_name)
        for param_name in STACK_QUERY_PARAMETERS:
            assert param_name in resource._global_parameters
            assert resource._global_parameters[param_name] == STACK_QUERY_PARAMETER_DEFAULT_VALUE





ALL_APIS = [('abusiveexperiencereport', 'v1'),
 ('acceleratedmobilepageurl', 'v1'),
 ('accesscontextmanager', 'v1beta'),
 ('adexchangebuyer', 'v1.4'),
 ('adexchangebuyer2', 'v2beta1'),
 ('adexperiencereport', 'v1'),
 ('admin', 'reports_v1'),
 ('adsense', 'v1.4'),
 ('adsensehost', 'v4.1'),
 ('alertcenter', 'v1beta1'),
 ('analytics', 'v3'),
 ('analyticsreporting', 'v4'),
 ('androiddeviceprovisioning', 'v1'),
 ('androidenterprise', 'v1'),
 ('androidmanagement', 'v1'),
 ('androidpublisher', 'v3'),
 ('appengine', 'v1'),
 ('appsactivity', 'v1'),
 ('appstate', 'v1'),
 ('bigquery', 'v2'),
 ('bigquerydatatransfer', 'v1'),
 ('binaryauthorization', 'v1beta1'),
 ('blogger', 'v3'),
 ('books', 'v1'),
 ('calendar', 'v3'),
 ('chat', 'v1'),
 ('civicinfo', 'v2'),
 ('classroom', 'v1'),
 ('cloudasset', 'v1beta1'),
 ('cloudbilling', 'v1'),
 ('cloudbuild', 'v1'),
 ('clouddebugger', 'v2'),
 ('clouderrorreporting', 'v1beta1'),
 ('cloudfunctions', 'v1'),
 ('cloudiot', 'v1'),
 ('cloudkms', 'v1'),
 ('cloudprofiler', 'v2'),
 ('cloudresourcemanager', 'v2'),
 ('cloudscheduler', 'v1beta1'),
 ('cloudsearch', 'v1'),
 ('cloudshell', 'v1'),
 ('cloudtasks', 'v2beta3'),
 ('cloudtrace', 'v2'),
 ('composer', 'v1'),
 ('compute', 'v1'),
 ('container', 'v1'),
 ('content', 'v2'),
 ('customsearch', 'v1'),
 ('dataflow', 'v1b3'),
 ('dataproc', 'v1'),
 ('datastore', 'v1'),
 ('deploymentmanager', 'v2'),
 ('dfareporting', 'v3.2'),
 ('dialogflow', 'v2'),
 ('digitalassetlinks', 'v1'),
 ('discovery', 'v1'),
 ('dlp', 'v2'),
 ('dns', 'v1'),
 ('doubleclickbidmanager', 'v1'),
 ('doubleclicksearch', 'v2'),
 ('drive', 'v3'),
 ('driveactivity', 'v2'),
 ('file', 'v1beta1'),
 ('firebasedynamiclinks', 'v1'),
 ('firebasehosting', 'v1beta1'),
 ('firebaserules', 'v1'),
 ('firestore', 'v1'),
 ('fitness', 'v1'),
 ('fusiontables', 'v2'),
 ('games', 'v1'),
 ('gamesConfiguration', 'v1configuration'),
 ('gamesManagement', 'v1management'),
 ('genomics', 'v1'),
 ('gmail', 'v1'),
 ('groupsmigration', 'v1'),
 ('groupssettings', 'v1'),
 ('iam', 'v1'),
 ('iamcredentials', 'v1'),
 ('iap', 'v1beta1'),
 ('identitytoolkit', 'v3'),
 ('indexing', 'v3'),
 ('jobs', 'v3'),
 ('kgsearch', 'v1'),
 ('language', 'v1'),
 ('licensing', 'v1'),
 ('logging', 'v2'),
 ('manufacturers', 'v1'),
 ('mirror', 'v1'),
 ('ml', 'v1'),
 ('monitoring', 'v3'),
 ('oauth2', 'v2'),
 ('oslogin', 'v1'),
 ('pagespeedonline', 'v5'),
 #('partners', 'v2'),
 ('people', 'v1'),
 ('playcustomapp', 'v1'),
 ('plus', 'v1'),
 ('plusDomains', 'v1'),
 ('poly', 'v1'),
 ('proximitybeacon', 'v1beta1'),
 ('pubsub', 'v1'),
 ('redis', 'v1'),
 ('replicapool', 'v1beta1'),
 ('replicapoolupdater', 'v1beta1'),
 ('reseller', 'v1'),
 ('runtimeconfig', 'v1'),
 ('safebrowsing', 'v4'),
 ('script', 'v1'),
 ('searchconsole', 'v1'),
 ('servicebroker', 'v1'),
 ('serviceconsumermanagement', 'v1'),
 ('servicecontrol', 'v1'),
 ('servicemanagement', 'v1'),
 ('servicenetworking', 'v1beta'),
 ('serviceusage', 'v1'),
 ('sheets', 'v4'),
 ('siteVerification', 'v1'),
 ('slides', 'v1'),
 ('sourcerepo', 'v1'),
 ('spanner', 'v1'),
 ('speech', 'v1'),
 ('sqladmin', 'v1beta4'),
 ('storage', 'v1'),
 ('storagetransfer', 'v1'),
 ('streetviewpublish', 'v1'),
 ('surveys', 'v2'),
 ('tagmanager', 'v2'),
 ('tasks', 'v1'),
 ('testing', 'v1'),
 ('texttospeech', 'v1'),
 ('toolresults', 'v1beta3'),
 ('tpu', 'v1'),
 ('translate', 'v2'),
 ('urlshortener', 'v1'),
 ('vault', 'v1'),
 ('videointelligence', 'v1'),
 ('vision', 'v1'),
 ('webfonts', 'v1'),
 ('webmasters', 'v3'),
 ('websecurityscanner', 'v1alpha'),
 ('youtube', 'v3'),
 ('youtubeAnalytics', 'v2'),
 ('youtubereporting', 'v1')]


import pytest

from aiogoogle import Aiogoogle
from aiogoogle.resource import Resource, GoogleAPI, Method, STACK_QUERY_PARAMETER_DEFAULT_VALUE, STACK_QUERY_PARAMETERS
from ..test_globals import ALL_APIS


@pytest.mark.parametrize('name,version', ALL_APIS)
def test_getitem(open_discovery_document, name, version):
    discovery_document = open_discovery_document(name, version)
    api = GoogleAPI(discovery_document=discovery_document)
    assert api['name'] == discovery_document.get('name')
    assert api['version'] == discovery_document.get('version')
    assert api['auth'] == discovery_document.get('auth')

@pytest.mark.parametrize('name,version', ALL_APIS)
def test_properties(open_discovery_document, name, version):
    discovery_document = open_discovery_document(name, version)
    api = GoogleAPI(discovery_document=discovery_document)
    assert (name in str(api)) or (discovery_document.get('title') in str(api))
    assert (version in str(api)) or (discovery_document.get('title') in str(api))

@pytest.mark.parametrize('name,version', ALL_APIS)
def test_constructor(open_discovery_document, name, version):
    discovery_document = open_discovery_document(name, version)
    api = GoogleAPI(discovery_document=discovery_document)
    assert id(api.discovery_document) == id(api.discovery_document)
    assert api['resources'] == discovery_document.get('resources')

@pytest.mark.parametrize('name,version', ALL_APIS)
def test_getattr(open_discovery_document, name, version):
    discovery_document = open_discovery_document(name, version)
    api = GoogleAPI(discovery_document=discovery_document)
    for resource_name, _ in discovery_document.get('resources').items():
        resource_object = getattr(api, resource_name)
        assert isinstance(resource_object, Resource)
        assert resource_name == resource_object.name

@pytest.mark.parametrize('name,version', ALL_APIS)
def test_repr(open_discovery_document, name, version):
    discovery_document = open_discovery_document(name, version)
    api = GoogleAPI(discovery_document=discovery_document)
    name = discovery_document.get('name')
    version = discovery_document.get('version')
    base_url = discovery_document.get('baseUrl')
    assert name in str(api) and version in str(api) and base_url in str(api)

@pytest.mark.parametrize('name,version', ALL_APIS)
def test_len(open_discovery_document, name, version):
	discovery_document = open_discovery_document(name, version)
	api = GoogleAPI(discovery_document=discovery_document)
	methods_len = len(discovery_document.get('methods')) if discovery_document.get('methods') else 0
	resources_len = len(discovery_document.get('resources')) if discovery_document.get('resources') else 0
	assert len(api) == methods_len + resources_len

@pytest.mark.parametrize('name,version', ALL_APIS)
def test_resources_getattr_fails_on_unknown_resource(open_discovery_document, name, version):
    discovery_document = open_discovery_document(name, version)
    api = GoogleAPI(discovery_document=discovery_document)
    nonexistent_resource = 'I_MUST_NEVER_EXIST_ANYWHEREEEEEEEE'
    with pytest.raises(AttributeError) as e:
        getattr(api, nonexistent_resource)

@pytest.mark.parametrize('name,version', ALL_APIS)
def test_add_global_params(open_discovery_document, name, version):
    discovery_document = open_discovery_document(name, version)
    api = GoogleAPI(discovery_document=discovery_document)
    for param in STACK_QUERY_PARAMETERS:
        assert param in api['parameters']
        assert api['parameters'][param] == STACK_QUERY_PARAMETER_DEFAULT_VALUE



__all__ = [
    'client',
    'excs',
    'resource',
    'utils',
    'validate'
]

from .client import Aiogoogle
from .resource import GoogleAPI
from .excs import AiogoogleError, AuthError, HTTPError, ValidationError



'''
A simple instance validation module for Discovery schemas.
Unfrtunately, Google uses a slightly modified version of JSONschema draft3 (mix of jsonschema and protobuff).
As a result, using an external library to validate Discovery schemas will raise lots of errors.
I tried to modify the popular: https://github.com/Julian/jsonschema to make it work with Google's version,
but it was just too complicated for the relatively simple task on our hands.
If you face any problems with aiogoogle's validation, you can always turn it off by either passing: ``method.__call__(validate=False)``
or turning it off for the whole API by passing aiogoogle.discover(validate=False)

This module misses a lot of the features provided with more advanced jsonschema validators e.g.

1. collecting all validation errors and raising them all at once
2. more descriptive errors with nice tracebacks
'''


__all__ = [
    'validate',
]

import datetime
import warnings
import re
import rfc3339

from .excs import ValidationError


#------- MAPPINGS -------#

JSON_PYTHON_TYPE_MAPPING = {
    'number': (float, int),
    'integer': (int),
    'string': (str),
    'object': (dict),
    'array': (list, set, tuple),
    'boolean': (bool),
    'null': (),  # Not used
    'any': (float, int, str, dict, list, set, tuple, bool)
}

KNOWN_FORMATS = ['int32', 'uint32', 'double', 'float', 'null', 'byte', 'date', 'date-time', 'int64', 'uint64', 'google-fieldmask', 'google-duration', 'google-datetime']

IGNORABLE_FORMATS = ['google-fieldmask', 'google-duration', 'google-datetime']  # After writing a format check for any of those formats, remove the format you wrote the checker for.

TYPE_FORMAT_MAPPING = {
    # Given this type, if None then don't check for additional "format" property in the schema, else, format might be any of the mapped values
    # Those are kept here for reference. They aren't used by any of the validators below. Instead validators check directly if there's any format requirements
    'any': [],
    'array': [],
    'boolean': [],
    'integer': ['int32', 'unint32'],
    'number': ['double', 'float'],
    'object': [],
    'string': ['null', 'byte', 'date', 'date-time','int64','uint64']
}

#------ Helpers -------#

def remove_microseconds(value):
    # 2014-02-11T14:13:26
    if not isinstance(value, str):
        raise ValidationError("datetime should be a string")
    if '.' in value:
        print(value)
        print(value[:-5])
        return value[:-5] + 'Z'
    else:
        return value

def make_validation_error(checked_value, correct_criteria):
    return f"{checked_value} isn't valid. Expected a value that meets the following criteria: {correct_criteria}"

#-------- VALIDATORS ---------#

# Type validators (JSON schema)

def any_validator(value):
    req_types = JSON_PYTHON_TYPE_MAPPING['any']
    if not isinstance(value, req_types):
        raise ValidationError(make_validation_error(value, str(req_types)))

def array_validator(value):
    req_types = JSON_PYTHON_TYPE_MAPPING['array']
    if not isinstance(value, req_types):
        raise ValidationError(make_validation_error(value, str(req_types)))

def boolean_validator(value):
    req_types = JSON_PYTHON_TYPE_MAPPING['boolean']
    if not isinstance(value, req_types):
        raise ValidationError(make_validation_error(value, str(req_types)))

def integer_validator(value):
    req_types = JSON_PYTHON_TYPE_MAPPING['integer']
    if not isinstance(value, req_types):
        raise ValidationError(make_validation_error(value, str(req_types)))

def number_valdator(value):
    req_types = JSON_PYTHON_TYPE_MAPPING['number']
    if not isinstance(value, req_types):
        raise ValidationError(make_validation_error(value, str(req_types)))

def object_validator(value):
    req_types = JSON_PYTHON_TYPE_MAPPING['object']
    if not isinstance(value, req_types):
        raise ValidationError(make_validation_error(value, str(req_types)))

def string_validator(value):
    req_types = JSON_PYTHON_TYPE_MAPPING['string']
    if not isinstance(value, req_types):
        raise ValidationError(make_validation_error(value, str(req_types)))

# Format validators (Discovery Specific)  https://developers.google.com/discovery/v1/type-format

def int32_validator(value):
    if (value > 2147483648) or (value < -2147483648) :
        raise ValidationError(make_validation_error(value, 'Integer between -2147483648 and 2147483648'))

def uint32_validator(value):
    if (value < 0) or (value > 4294967295):
        raise ValidationError(make_validation_error(value, 'Integer between 0 and 4294967295'))

def int64_validator(value):
    if (value > 9223372036854775807) or (value < -9223372036854775807):
        raise ValidationError(make_validation_error(value, 'Integer between -9,223,372,036,854,775,807 and -9,223,372,036,854,775,807'))

def uint64_validator(value):
    if (value > 9223372036854775807*2) or (value < 0):
        raise ValidationError(make_validation_error(value, 'Integer between 0 and 9,223,372,036,854,775,807 * 2'))

def double_validator(value):
    if not isinstance(value, float):
        raise ValidationError(make_validation_error(value, 'Double type'))

def float_validator(value):
    if not isinstance(value, float):
        raise ValidationError(make_validation_error(value, 'Float type'))

def byte_validator(value):
    if not isinstance(value, bytes):
        raise ValidationError(make_validation_error(value, 'Bytes type'))

def date_validator(value):
    msg = make_validation_error(value, 'JSON date value. Hint: use datetime.date.isoformat(), instead of datetime.date')
    try:
        pvalue = rfc3339.parse_date(value)
        #pvalue = datetime.date.fromisoformat(value)
    except Exception as e:
        raise ValidationError(str(e) + msg)
    if not isinstance(pvalue, datetime.date):
        raise ValidationError(msg)

def datetime_validator(value):
    msg = make_validation_error(value, 'JSON datetime value. Hint: use datetime.datetime.isoformat(), instead of datetime.datetime')
    try:
        pvalue = rfc3339.parse_datetime(value)
        #pvalue = datetime.datetime.fromisoformat(value)
    except Exception as e:
        raise ValidationError(str(e), msg)
    if not isinstance(pvalue, datetime.datetime):
        raise ValidationError(msg)

def null_validator(value):
    if value != 'null':
        raise ValidationError(make_validation_error(value, "'null' NOT None"))

# Other Validators

def minimum_validator(value, minimum):
    if value < int(minimum):
        raise ValidationError(make_validation_error(value, f'Not less than {minimum}'))

def maximum_validator(value, maximum):
    if value > int(maximum):
        raise ValidationError(make_validation_error(value, f'Not less than {maximum}'))

#-- Sub validators ---------------------------

def validate_type(instance, schema):
    type_validator_name = schema['type']
    # Check if type in list of possible types to avoid calling globals() maliciously
    if type_validator_name not in JSON_PYTHON_TYPE_MAPPING:
        warnings.warn(
            f"""\n\nUnknown type: {type_validator_name} found.\n\nSkipping type checks for {instance}
            against this schema:\n\n{schema}\n\nPlease create an 
            issue @ https://github.com/omarryhan/aiogoogle and report this warning msg.""")
        return
    type_validator = globals()[type_validator_name + '_validator']
    type_validator(instance)

def validate_format(instance, schema):
    if schema.get('format'):
        # Exception for format: mostly protobuf formats
        if schema['format'] in IGNORABLE_FORMATS:
            return
        # /Exception
        format_validator_name = schema['format']
        # Check if type in list of possible types to avoid calling globals() maliciously
        if format_validator_name not in KNOWN_FORMATS:
            warnings.warn(
            f"""\n\nUnknown format: {format_validator_name} found.\n\nSkipping format checks for {instance}
            against this schema:\n\n{schema}\n\nPlease create an 
            issue @ https://github.com/omarryhan/aiogoogle and report this warning msg.""")
            return
        if format_validator_name == 'date-time':
            format_validator_name = 'datetime'
        format_validator = globals()[format_validator_name + '_validator']
        format_validator(instance)

def validate_range(instance, schema):
    if schema.get('minimum'):
        minimum_validator(instance, schema['minimum'])
    elif schema.get('maximum'):
        maximum_validator(instance, schema['maximum'])

def validate_pattern(instance, schema):
    pattern = schema.get('pattern')
    if pattern is not None:
        match = re.match(pattern, instance)
        if match is None:
            raise ValidationError(instance, f'Match this pattern: {pattern}')

#-- Main Validator ---------------

def validate_all(instance, schema):
    validate_type(instance, schema)
    validate_format(instance, schema)
    validate_range(instance, schema)
    validate_pattern(instance, schema)

#-- API --------------------

def validate(instance, schema, schemas=None):
    '''
    Arguments:

        Instance: Instance to validate

        schema: schema to validate instance against

        schemas: Full schamas dict to resolve refs if any
    '''

    def resolve(schema):
        '''
        Resolves schema from schemas
        if no $ref was found, returns original schema
        '''
        if '$ref' in schema:
            if schemas is None:
                raise ValidationError(f"Attempted to resolve a {str(schema)}, but no schemas ref were found to resolve from")
            try:
                schema = schemas[schema['$ref']]
            except KeyError:
                raise ValidationError(f"Attempted to resolve {schema['$ref']}, but no results found.")
        return schema

    def validate_object():
        # Validate instance is an object
        object_validator(instance)  # dict validator, nothing jsonschema related

        # 1. Resolve additional properties
        additional_properties = schema.get('additionalProperties')
        # Jsonschema draft 3
        if additional_properties is False or (isinstance(additional_properties, dict) and len(additional_properties) ==0):
            additional_properties = None
        if not isinstance(additional_properties, dict): 
            # Typically, additionalProperties should be either False, dict(schema) or empty.
            # Empty will defailt to None, and False, as shown above, also defaults to None.
            # Sometimes, it's a str. Strings shouldn't be schemas. Strings will be ignored. Bad strings..
            # There's just too many of them to raise an error whenever we find one (around 75). So they'l just be ignored
            #raise ValidationError(f'Invalid type of addiotional properties. Shoudl be either a dict or False')
            additional_properties = None

        # 2. Resolve properties
        if not schema.get('properties'):
            if not additional_properties:
                raise ValidationError(f'''
                    Invalid Schema: {str(schema)}. 
                    Neither properties nor addiotional properties found in this schema'''
                )
            props = {}
        else:
            props = schema['properties']

        # 3. Raise warnings or fail on passed dict keys that aren't mentioned in the schema
        for k, _ in instance.items():
            if k not in props:
                # If there's a schema for additional properties validate
                if additional_properties is not None:
                    # some additional properties are misused and 
                    # have many properties instead of just one.
                    # Basically, additionalProperties is sort of like **kwargs, 
                    # you should only use one per statement/schema
                    if 'properties' not in additional_properties:
                        validate(instance[k], additional_properties)  # instace, schema for any additional props
                else:
                    warnings.warn(f"""
                        Item {k} was passed, but not mentioned in the
                        following schema {schema.get('id')}.\n\n
                        It will probably be discarded by the API you're using"""
                    )
        
        # 4. Validate existing properties
        for k,v in props.items():
            # Check if there's a ref to resolve
            v = resolve(v)
            # Check if instance has the property, if not, check if it's required
            if k not in instance:
                if v.get('required') is True:
                    raise ValidationError(f"Instance {k} is required")
            else:
                validate(instance[k], v, schemas)

    def validate_array():
        array_validator(instance)
        schema_ = resolve(schema['items'])
        # Check if instance has the property, if not, check if it's required
        for item in instance:
            validate(item, schema_, schemas)

    # Check schema and schemas are dicts. 
    # These errors shouldn't normally be raised, unless there's some messed up schema(s) being passed
    if not isinstance(schema, dict):
        raise TypeError('Schema must be a dict')
    if schemas is not None:
        if not isinstance(schemas, dict):
            raise TypeError('Schemas must be a dict')
    # Preliminary resolvement. 
    schema = resolve(schema)
    # If object (Dict): iterate over each entry and recursively validate
    if schema['type'] == 'object':
        validate_object()
    # If array (list or tuple) iterate over each item and recursively validate
    elif schema['type'] == 'array':
        # Validate instance is an array
        validate_array()
    # Else we reached the lowest level of a schema, validate
    else:
        validate_all(instance, schema)



__all__ = [
    'TrioAsksSession'
]

import trio
import asks
from asks import Session
asks.init('trio')

from .abc import AbstractSession
from ..models import Response


class TrioAsksSession(Session, AbstractSession):
    def __init__(self, *args, **kwargs):
        if kwargs.get('timeout'):
            del kwargs['timeout']
            kwargs.pop('timeout', None)
        super().__init__(*args, **kwargs)

    async def __aiter__(self):
        return self

    async def __aexit__(self, *args):
        pass

    async def send(self, *requests, timeout=None, full_res=False, raise_for_status=True, session_factory=None):
        responses = []

        def call_callback(request, response):
            if request.callback is not None:
                if response.json:
                    response.json = request.callback(response.content)
                elif response.data:
                    response.data = request.callback(response.content)
            return response

        async def resolve_response(request, response):
            data = None
            json = None
            download_file = None
            upload_file = None

            # If downloading file:
            if request.media_download:
                raise NotImplementedError('Downloading media isn\'t supported by this session')
            else:
                if response.status_code != 204:  # If no (no content)
                    try:
                        json = response.json()
                    except:
                        try:
                            data = response.text
                        except:
                            try:
                                data = response.content
                            except:
                                try:
                                    data = response.body
                                except:
                                    data = None
            
            if request.media_upload:
                upload_file = request.media_upload.file_path

            return Response(
                url=str(response.url),
                headers=response.headers,
                status_code=response.status_code,
                json=json,
                data=data,
                reason=response.reason_phrase if getattr(response, 'reason_phrase') else None,
                req=request,
                download_file=download_file,
                upload_file=upload_file,
                session_factory=session_factory
            )

        async def fire_request(request):
            request.headers['Accept-Encoding'] = 'gzip'
            if request.media_upload:
                raise NotImplementedError('Uploading media isn\'t supported by this session')
            else:
                return await self.request(
                    method=request.method,
                    url=request.url,
                    headers=request.headers,
                    data=request.data,
                    json=request.json,
                )

        #----------------- runners ------------------#
        async def get_response(request):
            response = await fire_request(request)
            response = await resolve_response(request, response)
            if raise_for_status is True:
                response.raise_for_status()
            response = call_callback(request, response)
            responses.append(response)
        async def get_content(request):
            response = await fire_request(request)
            response = await resolve_response(request, response)
            if raise_for_status is True:
                response.raise_for_status()
            response = call_callback(request, response)
            responses.append(response.content)
        #----------------- /runners ------------------#

        if session_factory is None:
            session_factory = self.__class__

        async def execute_tasks():
            async with trio.open_nursery() as nursery:
                if full_res is True:
                    [nursery.start_soon(get_response, request) for request in requests]
                else:
                    [nursery.start_soon(get_content, request) for request in requests]

        if timeout is not None:
            with trio.move_on_after(timeout):
                await execute_tasks()
        else:
            await execute_tasks()
        if isinstance(responses, list) and len(responses) == 1:
            return responses[0]
        else:
            return responses



import pytest

from aiogoogle import Aiogoogle
from aiogoogle.resource import Resource, GoogleAPI, Method, STACK_QUERY_PARAMETER_DEFAULT_VALUE, STACK_QUERY_PARAMETERS
from ..test_globals import ALL_APIS


@pytest.mark.parametrize('name,version', ALL_APIS)
def test_parameters(open_discovery_document, name, version):
    discovery_document = open_discovery_document(name, version)
    api = GoogleAPI(discovery_document=discovery_document)
    for resource_name, _ in discovery_document.get('resources').items():
        resource = getattr(api, resource_name)
        for method_name in resource.methods_available:
            method = getattr(resource, method_name)
            for parameter_name, _ in method.parameters.items():
                assert (
                    parameter_name in discovery_document.get('parameters') or 
                    parameter_name in discovery_document['resources'][resource_name]['methods'][method_name]['parameters']
                ) 

@pytest.mark.parametrize('name,version', ALL_APIS)
def test_optional_parameters(open_discovery_document, name, version):
    discovery_document = open_discovery_document(name, version)
    api = GoogleAPI(discovery_document=discovery_document)
    for resource_name, _ in discovery_document.get('resources').items():
        resource = getattr(api, resource_name)
        for method_name in resource.methods_available:
            method = getattr(resource, method_name)
            for parameter_name in method.optional_parameters:
                parameter = method.parameters[parameter_name]
                assert parameter.get('required') is not True

@pytest.mark.parametrize('name,version', ALL_APIS)
def test_required_parameters(open_discovery_document, name, version):
    discovery_document = open_discovery_document(name, version)
    api = GoogleAPI(discovery_document=discovery_document)
    for resource_name, _ in discovery_document.get('resources').items():
        resource = getattr(api, resource_name)
        for method_name in resource.methods_available:
            method = getattr(resource, method_name)
            for parameter_name in method.required_parameters:
                parameter = method.parameters[parameter_name]
                assert parameter.get('required') is True

@pytest.mark.parametrize('name,version', ALL_APIS)
def test_path_parameters(open_discovery_document, name, version):
    discovery_document = open_discovery_document(name, version)
    api = GoogleAPI(discovery_document=discovery_document)
    for resource_name, _ in discovery_document.get('resources').items():
        resource = getattr(api, resource_name)
        for method_name in resource.methods_available:
            method = getattr(resource, method_name)
            for parameter_name in method.path_parameters:
                parameter = method.parameters[parameter_name]
                assert parameter.get('location') == 'path'

@pytest.mark.parametrize('name,version', ALL_APIS)
def test_query_parameters(open_discovery_document, name, version):
    discovery_document = open_discovery_document(name, version)
    api = GoogleAPI(discovery_document=discovery_document)
    for resource_name, _ in discovery_document.get('resources').items():
        resource = getattr(api, resource_name)
        for method_name in resource.methods_available:
            method = getattr(resource, method_name)
            for parameter_name in method.query_parameters:
                parameter = method.parameters[parameter_name]
                assert parameter.get('location') == 'query'

def test_getitem():
    method = Method(
        name='IRRELEVANT',
        method_specs={
            'am_i_here_1': True,
            'am_i_here_2': None
        },
        global_parameters={'IRRELEVANT': 'IRRELAVANT'},
        schemas={'IRRELEVANT': 'IRRELEVANT'},
        batch_path='IRRELEVANT',
        root_url='IRRELEVANT',
        service_path='IRRELEVANT',
        validate = False 
    )

    assert method['am_i_here_1'] is True
    assert method['am_i_here_2'] is None
    assert method['i_dont_exist'] is None

@pytest.mark.parametrize('name,version', ALL_APIS)
def test_len(open_discovery_document, name, version):
    discovery_document = open_discovery_document(name, version)
    api = GoogleAPI(discovery_document=discovery_document)
    for resource_name, _ in discovery_document.get('resources').items():
        resource = getattr(api, resource_name)
        for method_name in resource.methods_available:
            method = getattr(resource, method_name)
            assert isinstance(len(method), int)

@pytest.mark.parametrize('name,version', ALL_APIS)
def test_str_repr(open_discovery_document, name, version):
    discovery_document = open_discovery_document(name, version)
    api = GoogleAPI(discovery_document=discovery_document)
    for resource_name, _ in discovery_document.get('resources').items():
        resource = getattr(api, resource_name)
        for method_name in resource.methods_available:
            method = getattr(resource, method_name)
            assert str(method)

@pytest.mark.parametrize('name,version', ALL_APIS)
def test_stack_parameters(open_discovery_document, name, version):
    discovery_document = open_discovery_document(name, version)
    api = GoogleAPI(discovery_document=discovery_document)
    for resource_name, _ in discovery_document.get('resources').items():
        resource = getattr(api, resource_name)
        for method_name in resource.methods_available:
            method = getattr(resource, method_name)
            for stack_param_name in STACK_QUERY_PARAMETERS:
                assert stack_param_name in method.optional_parameters
                assert stack_param_name in method.parameters



__all__ = []

from functools import wraps
from .models import Request, Response

def _toggle2x_dashed_params(f):
    @wraps(f)
    def wrapper(self, validate=None, data=None, json=None, upload_file=None, 
                download_file=None, timeout=None, **uri_params):
        '''
        Momentarily adds back '-' to url parameters and passed uri_params
        in order to be processed correctly and comply with the disc doc
        Reverts back to '_' after wrapped function is done
        '''
        # unfix urls
        uri_params = self._add_dash_user_uri_params(uri_params)
        
        # unfix params
        self._method_specs['parameters'] = self._add_dash_params(self._method_specs.get('parameters'))
        self._global_parameters = self._add_dash_params(self._global_parameters)

        # Run function
        results = f(self, validate, data, json, upload_file, download_file, timeout, **uri_params)

        # fix params again
        self._method_specs['parameters'] = self._rm_dash_params(self._method_specs.get('parameters'))
        self._global_parameters = self._rm_dash_params(self._global_parameters)

        return results
    return wrapper

def _safe_getitem(dct, *keys):
    for key in keys:
        try:
            dct = dct[key]
        except (KeyError):
            return None
    return dct

class _dict(dict):  # pragma: no cover
    ''' A simple dict subclass. Nothing special '''
    def __init__(self, *args, **kwargs):  # pragma: no cover
        super(_dict, self).__init__(*args, **kwargs)
        for arg in args:
            if isinstance(arg, dict):
                for k, v in arg.items():
                    self[k] = v

        if kwargs:
            for k, v in kwargs.items():
                self[k] = v

    def __getattr__(self, attr):  # pragma: no cover
        return self.get(attr)

    def __setattr__(self, key, value):  # pragma: no cover
        self.__setitem__(key, value)

    def __setitem__(self, key, value):  # pragma: no cover
        super(_dict, self).__setitem__(key, value)
        self.__dict__.update({key: value})

    def __delattr__(self, item):  # pragma: no cover
        self.__delitem__(item)

    def __delitem__(self, key):  # pragma: no cover
        super(_dict, self).__delitem__(key)
        del self.__dict__[key]



__all__ = [
    'ApiKey',
    'UserCreds',
    'ClientCreds'
]


from ..utils import _dict


class ApiKey(str):
    pass

class IdToken(_dict):
    '''
    OpenID connect ID token

    Example:

        {"iss":"accounts.google.com",
        "at_hash":"HK6E_P6Dh8Y93mRNtsDB1Q",
        "email_verified":"true",
        "sub":"10769150350006150715113082367",
        "azp":"1234987819200.apps.googleusercontent.com",
        "email":"jsmith@example.com",
        "aud":"1234987819200.apps.googleusercontent.com",
        "iat":1353601026,
        "exp":1353604926,
        "nonce": "0394852-3190485-2490358",
        "hd":"example.com"}

    Attributes:

        iss (str):

            * PROVIDED: always
        
            * The Issuer Identifier for the Issuer of the response. Always https://accounts.google.com or accounts.google.com for Google ID tokens.
        
        at_hash	(str):
        
            * Access token hash.
            
            * Provides validation that the access token is tied to the identity token.
            
            * If the ID token is issued with an access token in the server flow, this is always included.
            
            * This can be used as an alternate mechanism to protect against cross-site request forgery attacks, but if you follow Step 1 and Step 3 it is not necessary to verify the access token.
        
        email_verified (bool)
        
            * True if the user's e-mail address has been verified; otherwise false.
        
        sub (str):

            * PROVIDED: always
        
            * An identifier for the user, unique among all Google accounts and never reused.
            
            * A Google account can have multiple emails at different points in time, but the sub value is never changed.
            
            * Use sub within your application as the unique-identifier key for the user.
        
        azp	(str):
        
            * The client_id of the authorized presenter.
            
            * This claim is only needed when the party requesting the ID token is not the same as the audience of the ID token.
            
            * This may be the case at Google for hybrid apps where a web application and Android app have a different client_id but share the same project.
        
        email (str):
        
            * The user's email address.
            
            * This may not be unique and is not suitable for use as a primary key.
            
            * Provided only if your scope included the string "email".
        
        profile (str):
        
            * The URL of the user's profile page. Might be provided when:
        
                * The request scope included the string "profile"
        
                * The ID token is returned from a token refresh
        
            * When profile claims are present, you can use them to update your app's user records. 
            
            * Note that this claim is never guaranteed to be present.
        
        picture (str):
        
            * The URL of the user's profile picture. Might be provided when:
        
                * The request scope included the string "profile"
        
                * The ID token is returned from a token refresh
        
            * When picture claims are present, you can use them to update your app's user records.
            
            * Note that this claim is never guaranteed to be present.
        
        name (str):
        
            * The user's full name, in a displayable form. Might be provided when:
        
                * The request scope included the string "profile"
        
                * The ID token is returned from a token refresh
        
            * When name claims are present, you can use them to update your app's user records. Note that this claim is never guaranteed to be present.
        
        aud (str):
        
            * PROVIDED:	always
            
            * The audience that this ID token is intended for.
            
            * It must be one of the OAuth 2.0 client IDs of your application.
        
        iat (str):
        
            * PROVIDED: always
            
            * The time the ID token was issued, represented in Unix time (integer seconds).
        
        exp (str):
        
        	* PROVIDED: always
            
            * The time the ID token expires, represented in Unix time (integer seconds).
        
        nonce (str): 
        
            * The value of the nonce supplied by your app in the authentication request.
            
            * You should enforce protection against replay attacks by ensuring it is presented only once.
        
        hd (str):
        
            * The hosted G Suite domain of the user. Provided only if the user belongs to a hosted domain.
    '''
    pass

class UserCreds(_dict):
    '''
    OAuth2 User Credentials Dictionary

    Attributes:

        access_token (str): Access Token
        refresh_token (str): Refresh Token
        expires_in (int): seconds till expiry from creation
        expires_at (str): JSON datetime ISO 8601 expiry datetime
        scopes (list): list of scopes owned by access token
        
        id_token (aiogoogle.auth.creds.IdToken): Decoded OpenID JWT
        id_token_jwt (str): Encoded OpenID JWT
        
        token_type (str): Bearer
        token_uri (str): URI where this token was issued from
        token_info_uri (str): URI where one could get more info about this token
        revoke_uri (str): URI where this token should be revoked
        
    '''
    def __init__(
        self,
        access_token=None,
        refresh_token=None,
        expires_in=None,
        expires_at=None,
        scopes=None,
        
        id_token=None,
        id_token_jwt=None,
        
        token_type=None,
        token_uri=None,
        token_info_uri=None,
        revoke_uri=None,
    ):
        self.access_token = access_token
        self.refresh_token = refresh_token
        self.expires_in = expires_in
        self.expires_at = expires_at
        self.scopes = scopes
        
        self.id_token = id_token
        self.id_token_jwt = id_token_jwt
        
        self.token_type = token_type
        self.token_uri = token_uri
        self.token_info_uri = token_info_uri
        self.revoke_uri = revoke_uri

class ClientCreds(_dict):
    '''
    OAuth2 Client Credentials Dictionary

    Examples:

        Scopes: ['openid', 'email', 'https://www.googleapis.com/auth/youtube.force-ssl']

    Arguments:

        client_id (str): OAuth2 client ID
        client_secret (str): OAuth2 client secret
        scopes (list): List of scopes that this client should request from resource server
        redirect_uri (str): client's redirect uri
    '''
    def __init__(self, client_id=None, client_secret=None, scopes=None, redirect_uri=None):
        self.client_id = client_id
        self.client_secret = client_secret
        self.scopes = scopes
        self.redirect_uri = redirect_uri



__all__ = [
    'AiohttpSession'
]

import asyncio
from json import JSONDecodeError

from aiohttp import ClientSession
from aiohttp import ClientTimeout
from aiohttp.client_exceptions import ClientResponseError, ContentTypeError
import aiofiles
from aiofiles import os as async_os
import async_timeout

from ..models import Response
from .abc import AbstractSession
from ..excs import HTTPError, ValidationError



class AiohttpSession(ClientSession, AbstractSession):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    async def send(self, *requests, timeout=None, full_res=False, raise_for_status=True, session_factory=None):
        def call_callback(request, response):
            if request.callback is not None:
                if response.json:
                    response.json = request.callback(response.content)
                elif response.data:
                    response.data = request.callback(response.content)
            return response

        async def resolve_response(request, response):
            data = None
            json = None
            download_file = None
            upload_file = None

            # If downloading file:
            if request.media_download:
                download_file = request.media_download.file_path
                async with aiofiles.open(download_file, 'wb+') as f:
                    async for line in response.content:
                        await f.write(line)
            else:
                if response.status != 204:  # If no (no content)
                    try:
                        json = await response.json()
                    except (JSONDecodeError, ContentTypeError):
                        try:
                            data = await response.text()
                        except ContentTypeError:
                            try:
                                data = await response.read()
                            except ContentTypeError:
                                data = None
            
            if request.media_upload:
                upload_file = request.media_upload.file_path

            return Response(
                url=str(response.url),
                headers=response.headers,
                status_code=response.status,
                json=json,
                data=data,
                reason=response.reason if getattr(response, 'reason') else None,
                req=request,
                download_file=download_file,
                upload_file=upload_file,
                session_factory=session_factory
            )

        async def aiter_file(file_name, chunk_size):
            ''' Async file generator '''
            async with aiofiles.open(file_name, 'rb') as f:
                chunk = await f.read(chunk_size)
                while chunk:
                    yield chunk
                    chunk = await f.read(chunk_size)

        async def get_file_size(full_file_path):
            stat = await async_os.stat(full_file_path)
            return stat.st_size

        async def fire_request(request):
            request.headers['Accept-Encoding'] = 'gzip'
            if request.media_upload:
                # Validate
                if request.media_upload.validate is True and request.media_upload.max_size is not None:
                    size = await get_file_size(request.media_upload.file_path)
                    max_size = request.media_upload.max_size
                    if size > max_size:
                        raise ValidationError(
                            f'"{request.media_upload.file_path}" has a size of {size/1000}KB. Max upload size for this endpoint is: {max_size/1000}KB.'
                        )

                # If multipart pass a file async generator 
                if request.media_upload.multipart is True:
                    return await self.request(
                        method=request.method,
                        url=request.media_upload.upload_path,
                        headers=request.headers,
                        data=aiter_file(request.media_upload.file_path, request.media_upload.chunk_size),
                        json=request.json,
                        timeout=request.timeout
                    )
                # Else load file to memory and send
                else:
                    async with aiofiles.open(request.media_upload.file_path, 'rb') as file:
                        read_file = await file.read()
                        return await self.request(
                            method=request.method,
                            url=request.media_upload.upload_path,
                            headers=request.headers,
                            data=read_file,
                            json=request.json,
                            timeout=request.timeout
                    )
            else:
                return await self.request(
                    method = request.method,
                    url = request.url,
                    headers = request.headers,
                    data = request.data,
                    json = request.json,
                    timeout = request.timeout
                )

        #----------------- runners ------------------#
        async def get_response(request):
            response = await fire_request(request)
            response = await resolve_response(request, response)
            if raise_for_status is True:
                response.raise_for_status()
            response = call_callback(request, response)
            return response
        async def get_content(request):
            response = await get_response(request)
            return response.content
        #----------------- /runners ------------------#

        async def schedule_tasks():
            if full_res is True:
                tasks = [asyncio.create_task(get_response(request)) for request in requests]
            else:
                tasks = [asyncio.create_task(get_content(request)) for request in requests]
            return await asyncio.gather(*tasks, return_exceptions=False)

        if session_factory is None:
            session_factory = self.__class__

        if timeout is not None:
            async with async_timeout.timeout(timeout):
                results = await schedule_tasks()
        else:
            results = await schedule_tasks()

        if isinstance(results, list) and len(results) == 1:
            return results[0]
        else:
            return results



#!/usr/bin/python3.7

from test_globals import ALL_APIS
from aiogoogle import Aiogoogle
import os, sys
import errno
import json
import asyncio
from aiohttp import ClientSession
import pprint

async def refresh_disc_docs_json():

    file_errors = []

    # Create new .data/ dir if one doesn't exists
    current_dir = os.getcwd()
    if current_dir[-9:] != 'aiogoogle':  # current dir is aiogoogle
        print(current_dir)
        print("must be in aiogoogle's dir, not test dir")
        sys.exit()

    # Refresh ALL_APIS in tests/tests_globals.py
    ALL_APIS = []
    print('Refreshing ALL_APIS in tests_globals.py')
    async with ClientSession() as sess:
        apis_pref = await sess.get('https://www.googleapis.com/discovery/v1/apis?preferred=true')
        apis_pref = await apis_pref.json()
    for api in apis_pref['items']:
        ALL_APIS.append((api['name'], api['version']))
    with open('tests/test_globals.py', 'w') as f:
        f.write(f'ALL_APIS = {pprint.pformat(ALL_APIS)}')
        print('SUCCESS!')

    # Refresh discovery files in tests/data
    for name, version in ALL_APIS:
        aiogoogle = Aiogoogle()
        print(f'Downloading {name}-{version}')
        try:
            #  https://www.googleapis.com/discovery/v1/apis/partners/v2/rest 
            # raises 502 as of datetime.datetime(2018, 12, 1, 17, 46, 38, 39391)
            google_api = await aiogoogle.discover(name, version)
        except Exception as e:
            file_errors.append(
                {
                    f'{name}-{version}': str(e)
                }
            )
            continue

        data_dir_name = current_dir + '/tests/data/'
        try:
            if not os.path.exists(data_dir_name):
                os.makedirs(data_dir_name)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
        
        # Save discovery docuemnt as .json file to the newly created data dir
        file_name = current_dir + '/tests/data/' + name + '_' + version + '_discovery_doc.json'
        with open(file_name, 'w') as discovery_file:
            json.dump(google_api.discovery_document, discovery_file)
        print(f'saved {name}-{version} to {file_name}')
    
    print('Done')
    if file_errors:
        print(f'Errors found: {str(file_errors)}')
        sys.exit(1)

if __name__ == '__main__':
    asyncio.run(refresh_disc_docs_json())


# TODO: Test requesr add query param
# TODO: Test request rm query param
# TODO: Test Request.from response
# TODO: Test next page generator
# TODO: Test response.__call__ returns pagination gen
# TODO: Test next_page returns a valid request


import pytest

from aiogoogle import Aiogoogle
from ..test_globals import ALL_APIS
from aiogoogle.auth.managers import OPENID_CONFIGS_DISCOVERY_DOC_URL as openid_url
from aiogoogle.auth.data import WELLKNOWN_OPENID_CONFIGS as currunt_openid_configs
from aiohttp import ClientSession

@pytest.mark.asyncio
@pytest.mark.parametrize('name,version', ALL_APIS)
async def test_download_aiogoogle(name, version):
    aiogoogle = Aiogoogle()
    google_api = await aiogoogle.discover(name, version)
    assert google_api.discovery_document

@pytest.mark.asyncio
@pytest.mark.parametrize('name,version', ALL_APIS)
async def test_download_aiogoogle_without_version(name, version):
    aiogoogle = Aiogoogle()
    google_api = await aiogoogle.discover(name)
    assert google_api.discovery_document['name'] == name
    assert google_api.discovery_document['version'] == version

@pytest.mark.asyncio
async def test_oauth2_manager_api_is_latest_version():
    CURRENT_OAUTH2_API_VERSION = 2
    async with Aiogoogle() as google:
        oauth2_apis_list = await google.list_api('oauth2')
    apis = oauth2_apis_list['items']
    versions_available = [int(api['version'][-1:]) for api in apis]
    for version in versions_available:
        assert CURRENT_OAUTH2_API_VERSION >= version

@pytest.mark.asyncio
async def test_latest_openid_configs():
    async with ClientSession() as sess:
        new_configs_res = await sess.get(openid_url)
        new_configs = await new_configs_res.json()
        assert new_configs == currunt_openid_configs


import re

import pytest

from aiogoogle import Aiogoogle
from aiogoogle.resource import RESERVED_KEYWORDS, GoogleAPI
from aiogoogle.validate import KNOWN_FORMATS, JSON_PYTHON_TYPE_MAPPING
from ..test_globals import ALL_APIS


@pytest.mark.parametrize('name,version', ALL_APIS)
def test_parameters_not_included_twice(open_discovery_document, name, version):
    ''' 
    I was curious whether global parameters might have identical names
    with Method.parameters.
    Fails if similarities found
    '''
    discovery_document = open_discovery_document(name, version)
    google_api = GoogleAPI(discovery_document=discovery_document)
    for resource_name, _ in discovery_document.get('resources').items():
        resource = getattr(google_api, resource_name)
        for method_name in resource.methods_available:
            method = getattr(resource, method_name)
            local_params = method['parameters'] or {}
            for parameter_name, _ in local_params.items():
                assert parameter_name not in method._global_parameters


@pytest.mark.parametrize('name,version', ALL_APIS)
def test_parameters_not_colliding_with_google_api__call__(open_discovery_document, name, version):
    discovery_document = open_discovery_document(name, version)
    google_api = GoogleAPI(discovery_document=discovery_document)
    for resource_name, _ in discovery_document.get('resources').items():
        resource = getattr(google_api, resource_name)
        for method_name in resource.methods_available:
            method = getattr(resource, method_name)
            params = method.parameters
            for param_name, _ in params.items():
                assert param_name not in RESERVED_KEYWORDS


@pytest.mark.parametrize('name,version', ALL_APIS)
def test_parameters_not_colliding_with_google_api__call__fails(open_discovery_document, name, version):
    ''' asserts previous test catches collisions if any'''

    COLLIDING_RESERVED_KEYWORDS = ['alt', 'part']

    class CollisionError(Exception):
        pass
    
    def check_collision(param_name):
        if param_name in COLLIDING_RESERVED_KEYWORDS:
            raise CollisionError()

    discovery_document = open_discovery_document(name, version)
    google_api = GoogleAPI(discovery_document=discovery_document)
    for resource_name, _ in discovery_document.get('resources').items():
        resource = getattr(google_api, resource_name)
        for method_name in resource.methods_available:
            method = getattr(resource, method_name)
            params = method.parameters
            with pytest.raises(CollisionError):
                for param_name, _ in params.items():
                    check_collision(param_name)


@pytest.mark.parametrize('name,version', ALL_APIS)
def test_properties_only_in_objects_not_any_other_type(open_discovery_document, name, version):
    '''
    Disc doc sanity check.
    Checks if objects are the only google-jsonschema type that has a an attribute call "Properties"
    '''
    discovery_document = open_discovery_document(name, version)
    google_api = GoogleAPI(discovery_document=discovery_document)
    if google_api.discovery_document.get('schemas'):
        for _,v in google_api.discovery_document['schemas'].items():
            if v['type'] != 'object':
                assert not v.get('properties')

@pytest.mark.parametrize('name,version', ALL_APIS)
def test_items_must_be_in_array_type(open_discovery_document, name, version):
    ''' Test that all array types in disc docs have an attribute called "items".
    aiogoogle.validate.validate assumes all arrays has an "items" property and it will 
    fail if it didn't find one '''
    disc_doc = open_discovery_document(name, version)
    schemas = disc_doc['schemas']

    def resolve(schema):
        if '$ref' in schema:
            return schemas[schema['$ref']]
        else:
            return schema

    for _, schema in disc_doc['schemas'].items():
        if 'properties' in schema:  # 6 failures without this check
            for _, prop in schema['properties'].items():
                prop = resolve(prop)
                if prop['type'] == 'array':
                    assert 'items' in prop

@pytest.mark.parametrize('name,version', ALL_APIS)
def test_schemas_for_unknown_formats_and_types(open_discovery_document, name, version):
    ''' Iterates over all available disc docs and checks all schemas if they have any unknown types or 
    formats other than the ones mentioned in discovery service's documentation
    aiogoogle.validate.KNOWN_FORMATS and keys of aiogoogle.validate.JSON_PYTHON_TYPE_MAPPING '''
    disc_doc = open_discovery_document(name, version)
    schemas = disc_doc['schemas']
    resources = disc_doc['resources']

    def resolve(schema):
        if '$ref' in schema:
            return schemas[schema['$ref']]
        else:
            return schema

    # schemas
    for _, schema in schemas.items():
        if 'properties' in schema:  # 6 failures without this check
            for _, prop in schema['properties'].items():
                prop = resolve(prop)
                assert prop['type'] in JSON_PYTHON_TYPE_MAPPING
                if 'format' in prop:
                    assert prop['format'] in KNOWN_FORMATS

    # methods
    # Doesn't cover all methods as this assumes the following heierarchy api.resource.method and doesn't check for any other combination
    # However, this should cover a fair part of most APIs
    for _, resource in resources.items():
        if 'methods' in resource:
            for _, method in resource['methods'].items():
                if 'parameters' in method:
                    for _, param in method['parameters'].items():
                        param = resolve(param)
                        assert param['type'] in JSON_PYTHON_TYPE_MAPPING
                        if 'format' in param:
                            assert param['format'] in KNOWN_FORMATS
        

@pytest.mark.parametrize('name,version', ALL_APIS)
def test_parameters_are_not_objects(open_discovery_document, name, version):
    ''' 
    I was curious whether global parameters might have identical names
    with Method.parameters.
    Fails if similarities found
    '''
    discovery_document = open_discovery_document(name, version)
    google_api = GoogleAPI(discovery_document=discovery_document)
    for resource_name, _ in discovery_document.get('resources').items():
        resource = getattr(google_api, resource_name)
        for method_name in resource.methods_available:
            method = getattr(resource, method_name)
            if method.parameters:
                for _,v in method.parameters.items():
                    assert v.get('type') != 'object' 
                    assert not v.get('properties')

@pytest.mark.parametrize('name,version', ALL_APIS)
def test_resources_and_methods_dont_share_name(open_discovery_document, name, version):
    ''' Asserts no resources and methods that are on the same level share the same name.
    If a name was being shared, Aiogoogle will give precedence to the resource, rendering the method
    inaccessible '''
    def assert_no_similarity(resource):
        for resource_name in resource.resources_available:
            assert resource_name not in resource.methods_available
            assert_no_similarity(getattr(resource, resource_name))
    assert_no_similarity(GoogleAPI(open_discovery_document(name, version)))

@pytest.mark.parametrize('name,version', ALL_APIS)
def test_schemas_consistency(open_discovery_document, name, version):
    '''
    Tests the following:

        1. every schema has a type
        2. If an schema of type object doesn't have properties, then it must have additionalProperties
        3. addiotionalProperties should only contain a single schema
        4. All schemas should be objects (dicts)
        5. Tests all patterns in all schemas compile without errors

    Errors:
        
        1. 1 test fails because a schema that is of type: "object" neither has properties nor additionalProperties:
        
            - Youtube-v3, schemas, TokenPagination

        2. 2 schemas/properties doesn't have a type:

            1) pagespeedonline-v5

                Category groups = {
                    'additional_properties': {
                        ....
                        'properties':
                            'description': {
                                'description': 'An optional human readable description of the category group.',
                                'type': 'string'
                            },
                            'title': {
                                'description': 'The title of the category group.',
                                'type': 'string'
                            }
                        }
                    }
                }


            2) discovery-v1:

                RestDescription: {
                    id: "RestDescription",
                    type: "object",
                    properties: {
                        auth: {
                            type: "object",
                            description: "Authentication information.",
                            properties: {
                                oauth2: {
                                    type: "object",
                                    description: "OAuth 2.0 authentication information.",
                                    properties: {
                                        scopes: {
                                            type: "object",
                                            description: "Available OAuth 2.0 scopes.",
                                            additionalProperties: {
                                                type: "object",
                                                description: "The scope value.",
                                                properties: {
                                                    description: {
                                                        type: "string",                          Indentation is wrong here 
                                                        description: "Description of scope."     Indentation is wrong here
                                                        
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }

        3. The following is a less major inconsistency and has been commented out from the test below: 
        Some additionalProperties return a string instead of a schema or bool
        Those additional properties will be ignored by default:
            
            https://tools.ietf.org/html/draft-zyp-json-schema-03#section-5.4:
            5.4.  additionalProperties

            This attribute defines a schema for all properties that are not
            explicitly defined in an object type definition.  If specified, the
            value MUST be a schema or a boolean.  If false is provided, no
            additional properties are allowed beyond the properties defined in
            the schema.  The default value is an empty schema which allows any
            value for additional properties.


        I know that Google's version of jsonschema3 isn't supposed to be 100% compliant but the errors above make it really hard
        to perform the simplest of validation
     '''

    # Avoid infinite recursion e.g. 
    # https://www.googleapis.com/discovery/v1/apis/bigquery/v2/rest / schemas / QueryParameterType schema
    schemas_checked = []

    def resolve(schema, schemas):
        ''' Resolves schema if resolvable '''
        if '$ref' in schema:
            schema = schemas[schema['$ref']]
        return schema

    def methods_gen(resource):
        ''' Generates all methods available in a given resource/googleAPI '''
        for method in resource.methods_available:
            yield getattr(resource, method)
        for n_resource in resource.resources_available:
            yield from methods_gen(getattr(resource, n_resource))

    def validate(schema_name, schema, schemas):
        schema = resolve(schema, schemas)

        assert 'type' in schema

        if schema['type'] == 'object':
            if 'properties' not in schema:
                assert 'additionalProperties' in schema
            else:
                for name, prop in schema['properties'].items():
                    if name not in schemas_checked:
                        schemas_checked.append(name)
                        validate(name, prop, schemas)
            if schema.get('additionalProperties'):
                schema = schema['additionalProperties']
                if isinstance(schema, dict):
                    if 'properties' not in schema:  # correct
                        if schema not in schemas_checked:
                            schemas_checked.append(schema)
                            validate('additionalProperties', schema, schemas)
                #    # If many additionalProperties, (it should only be one schema for all the additional props), iterate over them (Causes two more errors)
                #    else:
                #        for k,v in schema.items():
                #            # Might cause an infinite loop
                #            validate(k, v, schemas)
                #elif isinstance(prop, bool):
                #    pass
                #else:
                #    #raise TypeError(
                #    #    f'The following additionalProperty should be either a schema or bool not "{prop}"')
                #    pass
                #    #for _ in range(100000):
                #    #    print(prop)
        else:
            # all schemas/subschemas should eventually reach this point for testing
            if schema.get('pattern'):
                re.compile(schema['pattern'])
            

    disc_doc = open_discovery_document(name, version)
    schemas = disc_doc['schemas']
    assert isinstance(schemas, dict)
    
    # Validate methods
    gapi = GoogleAPI(disc_doc)
    for method in methods_gen(gapi):
        for name, param in method.parameters.items():
            validate(name, param, schemas)

    # Validate schemas
    for name, schema in disc_doc['schemas'].items():
        if name not in schemas_checked:
            validate(name, schema, schemas)



import json
import os

import pytest

from aiogoogle import Aiogoogle
from aiogoogle.resource import Resource, GoogleAPI, Method
from aiogoogle.models import Request, MediaDownload, MediaUpload, ResumableUpload
from aiogoogle.excs import ValidationError


from ..test_globals import ALL_APIS


def test_NONE_uri_params_removed(create_api):
    youtube = create_api('youtube', 'v3')
    req = youtube.videos.list(part='snippet', alt=None)
    assert req.url == 'https://www.googleapis.com/youtube/v3/videos?part=snippet'

def test_url_query_extra_warns(create_api):
    youtube = create_api('youtube', 'v3')
    with pytest.warns(UserWarning, match='i_am_extra'):
        req = youtube.videos.list(part='snippet', i_am_extra=True, validate=False)
    assert req.url == 'https://www.googleapis.com/youtube/v3/videos?part=snippet&i_am_extra=True'

def test_url_query_multi_query_arg(create_api):
    youtube = create_api('youtube', 'v3')
    req = youtube.videos.list(part='snippet', chart='idontknowhatthisisbutitshouldpass', validate=False)
    assert req.url == 'https://www.googleapis.com/youtube/v3/videos?part=snippet&chart=idontknowhatthisisbutitshouldpass'
    
def test_url_query_multi_query_arg_ordered_by_insertion(create_api):
    youtube = create_api('youtube', 'v3')
    req = youtube.videos.list(chart='idontknowhatthisisbutitshouldpass', part='snippet', validate=False)
    assert req.url == 'https://www.googleapis.com/youtube/v3/videos?chart=idontknowhatthisisbutitshouldpass&part=snippet'

def test_url_path_args_proper(create_api):
    calendar = create_api('calendar', 'v3')
    # Required path parameters are: 'calendarId' and 'ruleId'
    req = calendar.acl.get(calendarId='uno', ruleId='dos', validate=False)
    assert req.url == 'https://www.googleapis.com/calendar/v3/calendars/uno/acl/dos'

def test_url_path_args_reversed(create_api):
    calendar = create_api('calendar', 'v3')
    # Should stay in the same order even if kwarg insertion was reversed
    req = calendar.acl.get(ruleId='dos', calendarId='uno', validate=False)
    assert req.url == 'https://www.googleapis.com/calendar/v3/calendars/uno/acl/dos'

def test_url_path_args_reversed_with_extra_arg(create_api):
    calendar = create_api('calendar', 'v3')
    with pytest.warns(UserWarning, match='i_am_extra'):
        req = calendar.acl.get(ruleId='dos', calendarId='uno', i_am_extra=1, validate=False)
    assert req.url == 'https://www.googleapis.com/calendar/v3/calendars/uno/acl/dos?i_am_extra=1'

def test_url_path_args_and_query_path(create_api):
    calendar = create_api('calendar', 'v3')
    req = calendar.events.get(eventId='dos', calendarId='uno', maxAttendees=1, validate=False)
    assert req.url == 'https://www.googleapis.com/calendar/v3/calendars/uno/events/dos?maxAttendees=1'

def test_url_path_args_and_query_path_extra_arg(create_api):
    calendar = create_api('calendar', 'v3')
    with pytest.warns(UserWarning, match='i_am_extra'):
        req = calendar.events.get(i_am_extra='asdasdasdasd', eventId='dos', calendarId='uno', maxAttendees=1, validate=False)
    assert req.url == 'https://www.googleapis.com/calendar/v3/calendars/uno/events/dos?maxAttendees=1&i_am_extra=asdasdasdasd'

def test_method(create_api):
    calendar = create_api('calendar', 'v3')
    req = calendar.events.get(eventId='dos', calendarId='uno', validate=False)
    assert req.method == 'GET'

def test_timeout(create_api):
    calendar = create_api('calendar', 'v3')
    req = calendar.calendarList.get(calendarId=123, timeout=4, validate=False)
    assert req.timeout == 4

def test_timeout_fails_on_not_int(create_api):
    calendar = create_api('calendar', 'v3')
    with pytest.raises(TypeError) as e:
        calendar.calendarList.get(calendarId=123, timeout=True, validate=False)
    assert 'int' in str(e)

    with pytest.raises(TypeError) as e:
        calendar.calendarList.get(calendarId=123, timeout=False, validate=False)
    assert 'int' in str(e)

    with pytest.raises(TypeError) as e:
        calendar.calendarList.get(calendarId=123, timeout='asas', validate=False)
    assert 'int' in str(e)

def test_json(create_api):
    JSON = {'etag': 'an_etag'}

    youtube = create_api('youtube', 'v3')
    req = youtube.videos.insert(
        autoLevels=True,
        part='snippet',
        json=JSON,
        validate=False
    )
    assert req.url == 'https://www.googleapis.com/youtube/v3/videos?autoLevels=True&part=snippet'
    assert req.json == JSON
    assert req.data is None

def test_data(create_api):
    DATA = {'etag': 'an_etag'}

    youtube = create_api('youtube', 'v3')
    req = youtube.videos.insert(
        autoLevels=True,
        part='snippet',
        data=DATA,
        validate=False
    )
    assert req.url == 'https://www.googleapis.com/youtube/v3/videos?autoLevels=True&part=snippet'
    assert req.data == DATA
    assert req.json is None

def test_data_str(create_api):
    DATA = 'asdasds'

    youtube = create_api('youtube', 'v3')
    req = youtube.videos.insert(
        autoLevels=True,
        part='snippet',
        data=DATA,
        validate=False
    )
    assert req.url == 'https://www.googleapis.com/youtube/v3/videos?autoLevels=True&part=snippet'
    assert req.data == DATA
    assert req.json is None

def test_data_and_json_fails(create_api):
    DATA = {'etag': 'an_etag'}
    JSON = {'etag2': 'an_etag'}

    youtube = create_api('youtube', 'v3')
    with pytest.raises(TypeError):
        youtube.videos.insert(
            autoLevels=True,
            part='snippet',
            data=DATA,
            json=JSON,
            validate=False
        )

def test_download_file(create_api):
    DOWNLOAD_FILE = '/home/omar/Documents/captions.txt'

    youtube = create_api('youtube', 'v3')
    req = youtube.captions.download(id='a_path', download_file=DOWNLOAD_FILE, validate=True)
    assert req.media_download.file_path == DOWNLOAD_FILE

def test_download_file_unsupported_method_fails(create_api):
    DOWNLOAD_FILE = '/home/omar/Documents/captions.txt'

    youtube = create_api('youtube', 'v3')
    with pytest.raises(ValidationError) as e:
        youtube.videos.list(part='irrelevant', download_file=DOWNLOAD_FILE, validate=True)
        assert 'download_file' in str(e)

def test_upload_file_name(create_api):
    UPLAOD_FILE_NAME = '/home/omar/Videos/video.mp4'

    youtube = create_api('youtube', 'v3')
    req = youtube.videos.insert(
        autoLevels=True,
        part='snippet',
        upload_file=UPLAOD_FILE_NAME,
        validate=False
    )
    assert req.url == 'https://www.googleapis.com/youtube/v3/videos?autoLevels=True&part=snippet'
    assert req.media_upload.file_path == UPLAOD_FILE_NAME

def test_upload_simple_path(create_api):
    UPLAOD_FILE_NAME = '/home/omar/Videos/video.mp4'

    youtube = create_api('youtube', 'v3')
    req = youtube.videos.insert(
        autoLevels=True,
        part='snippet',
        upload_file=UPLAOD_FILE_NAME,
        validate=False
    )
    assert req.media_upload.upload_path == 'https://www.googleapis.com/upload/youtube/v3/videos?autoLevels=True&part=snippet'

def test_upload_simple_multipart(create_api):
    UPLAOD_FILE_NAME = '/home/omar/Videos/video.mp4'

    youtube = create_api('youtube', 'v3')
    req = youtube.videos.insert(
        autoLevels=True,
        part='snippet',
        upload_file=UPLAOD_FILE_NAME,
        validate=False
    )
    assert req.media_upload.multipart is True

def test_upload_file_fails_unsupported_method(create_api):
    UPLAOD_FILE_NAME = '/home/omar/Videos/video.mp4'

    youtube = create_api('youtube', 'v3')
    with pytest.raises(ValidationError) as e:
        youtube.videos.list(
            part='snippet',
            upload_file=UPLAOD_FILE_NAME,
            validate=True
        )
        assert 'upload_file' in str(e)

def test_resumable_upload(create_api):
    RESUMABLE_UPLOAD_SPECS = {
        'multipart': True,
        'path': '/resumable/upload/example/v3/resource'
    }
    METHOD_SPECS = {
        'path': 'resource/',
        'httpMethod': 'GET',
        'parameters': {},
        'supportsMediaUpload': True,
        'mediaUpload': {
            'accept': [
                "*/*",
                "application/octet-stream",
                "text/xml"
            ],

            'maxSize': "100MB",
            
            'protocols': {
                'simple': {
                    'multipart': True,
                    'path': "resource"
                },
                'resumable': RESUMABLE_UPLOAD_SPECS
            },
        }
    }
    ROOT_URL = 'https://example.com/'
    SERVICE_PATH = 'example/v3/'
    BATCH_PATH = 'https://example.com/api/v1/batch'

    method = Method(name='upload', method_specs=METHOD_SPECS, global_parameters={}, schemas={}, batch_path=BATCH_PATH, service_path=SERVICE_PATH, root_url=ROOT_URL, validate=False)
    req = method(upload_file='/home/omar/resumable_file.file')
    assert req.media_upload.resumable
    assert req.media_upload.resumable.upload_path == 'https://example.com/resumable/upload/example/v3/resource/'
    assert req.media_upload.resumable.multipart is True

def test_unresumable_upload(create_api):
    METHOD_SPECS = {
        'path': 'resource/',
        'httpMethod': 'GET',
        'parameters': {},
        'supportsMediaUpload': True,
        'mediaUpload': {
            'accept': [
                "*/*",
                "application/octet-stream",
                "text/xml"
            ],

            'maxSize': "100MB",
            
            'protocols': {
                'simple': {
                    'multipart': True,
                    'path': "resource"
                }
            },
        }
    }
    ROOT_URL = 'https://example.com/'
    BATCH_PATH = 'https://example.com/api/v1/batch'
    SERVICE_PATH = 'service/'

    method = Method(name='upload', method_specs=METHOD_SPECS, global_parameters={}, schemas={}, batch_path=BATCH_PATH, root_url=ROOT_URL, service_path=SERVICE_PATH, validate=False)
    req = method(upload_file='/home/omar/resumable_file.file')
    assert req.media_upload.resumable is None

def test_fix_params(create_api):
    urlshortener = create_api('urlshortener', 'v1')
    # It is 'start-token' in the disc doc
    assert 'start_token' in urlshortener.url.list.parameters
    req = urlshortener.url.list(start_token='mehh')
    assert req.url == 'https://www.googleapis.com/urlshortener/v1/url/history?start-token=mehh'  # Back to 'start-token' instead of 'start_token'
    assert 'start_token' in urlshortener.url.list.parameters



__all__ = [
    'GoogleAPI',
    'Resource',
    'Method'
]

import re
import warnings
from urllib.parse import urlencode
from functools import wraps
import datetime

from .excs import ValidationError
from .utils import _dict, _safe_getitem, _toggle2x_dashed_params
from .excs import ValidationError
from .models import MediaDownload, MediaUpload, ResumableUpload, Request
from .validate import validate as validate__


# These are the hard-coded kwargs in Method.__call__
# They're used for testing whether those names will collide with any of the url parameters that are provided by any of the discovery docs.
# If collisions were to be found, that would mean that the user won't be able to pass a url_parameter that shares the same name with any of the RESERVED_KEYWORDS.
RESERVED_KEYWORDS = ['validate', 'data', 'json', 'upload_file', 'download_file', 'timeout']

# From: https://github.com/googleapis/google-api-python-client/blob/master/googleapiclient/discovery.py
# Parameters accepted by the stack, but not visible via discovery.
STACK_QUERY_PARAMETERS = frozenset(['trace', 'pp', 'strict'])
STACK_QUERY_PARAMETER_DEFAULT_VALUE = {
    'type': 'string',
    'location': 'query',
}

# TODO: etagRequired: {
#    type: "boolean",
#    description: "Whether this method requires an ETag to be specified. The ETag is sent as an HTTP If-Match or If-None-Match header."
#    }
# Note: etagRequired is only mentioned once in all of the discovery documents available from Google. (In discovery_service-v1. So, it isn't actually being used)


class Method:
    def __init__(self, name, method_specs, global_parameters, schemas, root_url, service_path, batch_path, validate):
        # Replaces '-'s with '_'s and preserve old names to revert back to them after this method is called
        global_parameters = self._rm_dash_params(global_parameters)
        method_specs['parameters'] = self._rm_dash_params(method_specs.get('parameters'))

        self.name = name
        self._method_specs = method_specs
        self._global_parameters = global_parameters
        self._schemas = schemas

        self._root_url = root_url
        self._service_path = service_path
        self._batch_path = batch_path
        
        # Not gonna use it, because it's useless. Maybe I'm wrong?
        if self['useMediaDownloadService'] is True and self['supportsMediaDownload'] is True:
            self._download_base_url = self._root_url + 'download/' + self._service_path
        else:
            self._download_base_url = None

        self._base_url = self._root_url + self._service_path
        self._batch_url = self._root_url + self._batch_path

        self._should_validate = validate

#---- Changes URL parameters with a "-" to "_" -----# 
# Depends on how you view it, but this section also changes robots with small mouths to robots with big mouths

    @staticmethod
    def _rm_dash_params(param_set) -> dict:
        if param_set:
            for name, schema in param_set.items():
                if '-' in name:
                    new_name = name.replace('-', '_')  # See?!
                    schema['orig_name'] = name
                    param_set[new_name] = schema
                    del param_set[name]
        return param_set

    @staticmethod
    def _add_dash_params(param_set) -> dict:
        if param_set:
            # list() forces the creation of a new copy in memory
            # to avoid having the dict size changing during iteration
            for name, schema in list(param_set.items()):
                if 'orig_name' in schema:
                    param_set[schema['orig_name']] = schema
                    del param_set[name]
        return param_set

    def _add_dash_user_uri_params(self, uri_params):
        for k,v in uri_params.items():
            if '_' in k:
                if k in self.parameters:
                    if 'orig_name' in self.parameters[k]:
                        uri_params[self.parameters[k]['orig_name']] = v
                        del uri_params[k]
        return uri_params

#---- / Changes URL parameters with a "-" to "_" -----#

    @property
    def request(self) -> dict:
        ''' Returns expected request body '''
        body = self['request']
        if body.get('$ref'):
            return self._schemas.get(body['$ref'])
        else:
            return body

    @property
    def response(self) -> dict:
        ''' Retruns expected response body '''
        body = self['response']
        if body.get('$ref'):
            return self._schemas.get(body['$ref'])
        else:
            return body

    @property
    def parameters(self) -> dict:
        ''' 
        Parameters property

        Returns:

            dict: All parameters that this method can take as described in the discovery document
        '''
        if not self._global_parameters and not self['parameters']:
            return {}
        elif not self._global_parameters:
            return self['parameters']
        elif not self['parameters']:
            return self._global_parameters
        else:
            return {**self['parameters'], **self._global_parameters}

    @property
    def optional_parameters(self) -> [str, str]:
        '''
        Optional Parameters

        Returns:

            list: List of the names of optional parameters this method takes
        '''
        return [parameter_name for parameter_name, parameter_info in self.parameters.items() if parameter_info.get('required') is not True] if self.parameters else []

    @property
    def required_parameters(self) -> [str, str]:
        '''
        Required Parameters

        Returns:

            list: List of the names of required parameters this method takes
        '''
        return [parameter_name for parameter_name, parameter_info in self.parameters.items() if parameter_info.get('required') is True] if self.parameters else []
    
    @property
    def path_parameters(self) -> [str, str]:
        '''
        Path Parameters

        Returns:

            list: List of the names of path parameters this method takes
        '''
        return [param_name for param_name, param_info in self.parameters.items() if param_info.get('location') == 'path'] if self.parameters else []

    @property
    def query_parameters(self) -> [str, str]:
        '''
        Query Parameters

        Returns:

            list: List of the names of Query parameters this method takes
        '''
        return [param_name for param_name, param_info in self.parameters.items() if param_info.get('location') == 'query'] if self.parameters else []

    @property
    def required_query_parameters(self) -> [str, str]:
        '''
        Required Query Parameters

        Returns:

            list: List of the names of required query parameters this method takes
        '''
        return [param_name for param_name in self.query_parameters if param_name in self.required_parameters] if self.query_parameters else []

    @property
    def optional_query_parameters(self) -> [str, str]:
        '''
        Optional Query Parameters

        Returns:

            list: List of the names of optional query parameters this method takes
        '''
        return [param_name for param_name in self.query_parameters if param_name in self.optional_parameters] if self.query_parameters else []
    
    def __getitem__(self, key):
        '''
        Examples:

            ::

                >>> self['description'] 
                
                "method description"

                >>> self['scopes']
                
                ['returns', 'scopes', 'required', 'by', 'this', 'method', 'in', 'a', 'list']

                >>> self['supportsMediaDownload']
                
                False

                >>> self['supportsMediaUpload']

                True

                >>> self['httpMethod']

                'GET'

        Hint:


            Using this method with ``scopes`` as an argument can be useful for incremental authorization. (Requesting scopes when needed. As opposed to requesting them at once)

            for more: https://developers.google.com/identity/protocols/OAuth2WebServer#incrementalAuth


        '''
        return self._method_specs.get(key)

    def _validate(self, instance, schema):
        return validate__(instance, schema, self._schemas)

    @staticmethod
    def _rm_none_params(uri_params):
        for k,v in list(uri_params.items()):
            if v is None:
                del uri_params[k]
        return uri_params

    @_toggle2x_dashed_params
    def __call__(self, validate=None, data=None, json=None, upload_file=None, 
                download_file=None, timeout=None, **uri_params) -> Request:
        ''' 
        Builds a request from this method

        Note:

            * When passing ``datetime.datetime or datetime.date`` pass them in json format.
            
            * Aiogoogle won't do that as it would be a big hassle to iterate over every item in ``*uri_params``, ``json`` and ``data``
            to check if there's any datetime objects.

            * Fortunately Python makes it really easy to achieve that. 
            
            * Instead of passing say ``datetime.datetime.utcnow()``, pass: ``datetime.datetime.utcnow().jsonformat()``

        Note:

            * All ``None`` values are ommited before sending to Google apis, if you want to explicitly pass a JSON null then pass it as ``"null"`` not ``None`` 

        Arguments:

            validate (bool): Overrides :param: aiogoogle.Aiogoole.validate if not None
            
            json (dict): Json body
            
            data (any): Data body (Bytes, text, www-url-form-encoded and others)
            
            upload_file (str): full path of file to upload
            
            download_file (str): full path of file to download to
            
            timeout (str): total timeout for this request
            
            **uri_params (dict): path and query, required and optional parameters

        Returns:

            aiogoogle.models.Request: An unsent request object
        '''
        # If collisions are found between the 'key' of **uri_params and explicit kwargs e.g. data, json etc., then 
        # priority will be given to explicit kwargs. With that said, it's not likely there will be any.
        # If you want to double check if there are any collisions,
        # you can append the API name and version you're using to tests.globals.ALL_APIS (only if they don't exist, otherwise, you shouldn't worry about collisions)
        # Then, run the unit tests and monitor: tests.test_discovery_document.test_parameters_not_colliding_with_google_api__call__ for failure

        #
        # NOTE: Use '_' instead of '-' when passing uri parameters that have a '-' in their names
        #

        # Remove params that are None
        uri_params = self._rm_none_params(uri_params)

        # Assert timeout is int
        if timeout is not None:
            if not isinstance(timeout, int) or type(timeout) == bool:
                raise TypeError('Timeouts can only be ints or None')

        # Resolve validation status
        if not isinstance(validate, bool):
            validate = self._should_validate

        base_url = self._base_url
        
        # Build full url minus query & fragment
        url = self._build_url(base_url=base_url, uri_params=uri_params, validate=validate)

        # Filter out query parameters from all uri_params that were passed to this method
        passed_query_params = {param_name: param_info for param_name, param_info in uri_params.items() if param_name in self.query_parameters}

        # Ensure all required query parameteters were passed
        for param in self.required_query_parameters:
            if param not in passed_query_params:
                raise ValidationError(f'Missing query parameter: \"{param}\"')

        # Validate url query params
        if validate is True:
            if passed_query_params:
                for param_name, passed_param in passed_query_params.items():
                    self._validate(passed_param, self.parameters[param_name])
        
        # Join query params
        if passed_query_params:
            uri = url + '?' + urlencode(passed_query_params)
        else:
            uri = url

        # Pop uri_params consumed
        for param_name, _ in passed_query_params.items():
            del uri_params[param_name]

        # Warn if not all uri_params were consumed/popped
        if uri_params:  # should be empty by now
            # If there's room for addtionalProperties, validate and add them to the URI
            if validate:
                if self.parameters.get('additionalProperties'):
                    for _,v in uri_params.items():
                        self._validate(v, self.parameters['additionalProperties'])
                else:
                    raise ValidationError('Invalid parameters: {} were passed')
            else:
                if not self.parameters.get('additionalProperties'):
                    warnings.warn('Parameters {} were found and they\'re probably of no use.'
                            ' Check if they\'re valid parameters'.format(str(uri_params)))
            if '?' not in uri:
                uri = uri + '?' + urlencode(uri_params)
            else:
                uri = uri + '&' + urlencode(uri_params)
                

        # Ensure only one param for http req body.
        if json and data:
            raise TypeError('Pass either json or data for the body of the request, not both.'
                            '\nThis is similar to the "body" argument in google-python-client'
                            '\nThis will validate agains the $request key in this method\'s '
                            'specs')  # This raises a TypeError instead of a ValidationError because 
                                      # it will probably make your session raise an error if it passes.
                                      # Better raise it early on

        # Validate body
        if validate is True:
            body = json if json is not None else data if data is not None else None  # json or data or None
            if body is not None:
                self._validate_body(body)

        # Process download_file
        if download_file:
            if validate is True:
                if self.__getitem__('supportsMediaDownload') is not True:
                    raise ValidationError('download_file was provided while method doesn\'t support media download')
            media_download = MediaDownload(download_file)
        else:
            media_download = None

        # Process upload_file
        if upload_file:
            media_upload = self._build_upload_media(upload_file, uri, validate, fallback_url=url)
        else:
            media_upload = None

        return Request(
            method=self['httpMethod'],
            url=uri,
            batch_url=self._batch_url,
            data=data,
            json=json,
            timeout=timeout,
            media_download=media_download,
            media_upload=media_upload,
            callback=lambda res: self._validate_response(res, validate)
        )

    def _build_url(self, base_url, uri_params, validate):
        if self.path_parameters:
            # sort path params as sepcified in method_specs.parameterOrder
            sorted_required_path_params = {}  # Dict order is guaranteed (by insertion) as of Python 3.6
            for param_name in self['parameterOrder']:
                try:
                    sorted_required_path_params[param_name] = uri_params.pop(param_name)
                except KeyError:
                    raise ValidationError(f'Missing URL path parameter: {param_name}')

            # Validate path params
            if validate is True:
                self._validate_url(sorted_required_path_params)

        # Build full path
            # replace named placeholders with empty ones. e.g. {param} --> {}
            # Why? Because some endpoints have different names in their url path placeholders than in their parameters
            # e.g. path: {"v1/{+resourceName}/connections"}. e.g. param name: resourceName NOT +resourceName
            self._method_specs['path'] = re.sub(r'\{(.*?)\}', r'{}', self._method_specs['path'])
            return base_url + self['path'].format(*list(sorted_required_path_params.values()))
        else:
            return base_url + self['path']

    def _build_upload_media(self, upload_file, qualified_url, validate, fallback_url):
        if self['supportsMediaUpload'] is not True:
            if validate is True:
                raise ValidationError('upload_file was provided while method doesn\'t support media upload')
            else:
                # This will probably not work, but will return a mediaupload object anyway
                return MediaUpload(upload_file, upload_path=fallback_url)
        
        # If resumable, create resumable object
        resumable = self._build_resumeable_media(upload_file, qualified_url) if _safe_getitem(self._method_specs, 'mediaUpload', 'protocols', 'resumable') else None
        
        # Create MediaUpload object and pass it the resumable object we just created
        #simple_upload_path = _safe_getitem(self._method_specs, 'mediaUpload', 'protocols', 'simple', 'path')
        #media_upload_url_base = self._root_url[:-1] + simple_upload_path
        media_upload_url_base = self._root_url + 'upload/' + self._service_path
        media_upload_url = qualified_url.replace(self._base_url, media_upload_url_base)
        max_size = self._convert_str_size_to_int(_safe_getitem(self._method_specs, 'mediaUpload', 'maxSize'))
        mime_range = _safe_getitem(self._method_specs, 'mediaUpload', 'accept')
        multipart = self['mediaUpload']['protocols']['simple'].get('multipart', True)

        # Return
        return MediaUpload(file_path=upload_file, upload_path=media_upload_url, max_size=max_size, mime_range=mime_range, multipart=multipart, resumable=resumable, validate=validate)

    def _build_resumeable_media(self, upload_file, qualified_url):
        #resumable_upload_path = _safe_getitem(self._method_specs, 'mediaUpload', 'protocols', 'resumable', 'path')
        #resumable_url_base = self._root_url[:-1] + resumable_upload_path
        resumable_url_base = self._root_url + 'resumable/upload/' + self._service_path
        resumable_url = qualified_url.replace(self._base_url, resumable_url_base)
        multipart = self['mediaUpload']['protocols']['resumable'].get('multipart', True)
        return ResumableUpload(upload_file, multipart=multipart, upload_path=resumable_url)

    @staticmethod
    def _convert_str_size_to_int(size):
        """Convert a string media size, such as 10GB or 3TB into an integer.

        Args:
            size: (str): e.g. such as 2MB or 7GB.

        Returns:
            The size as an integer value.
        """
        MEDIA_SIZE_BIT_SHIFTS = {'KB': 10, 'MB': 20, 'GB': 30, 'TB': 40}
        if size is None:
            return None
        if len(size) < 2:
            return 0
        units = size[-2:].upper()
        bit_shift = MEDIA_SIZE_BIT_SHIFTS.get(units)
        if bit_shift is not None:
            return int(size[:-2]) << bit_shift
        else:
            return int(size)

    def _validate_url(self, sorted_required_path_params):
        for path_param_name, path_param_info in sorted_required_path_params.items():
            self._validate(instance=path_param_info, schema=self.parameters[path_param_name])

    def _validate_body(self, req):
        request_schema = self._method_specs.get('request')
        if request_schema is not None:
            if '$ref' in request_schema:
                request_schema = self._schemas[request_schema['$ref']]
            self._validate(req, request_schema)
        else:
            raise ValidationError('Request body should\'ve been validated, but wasn\'t because the method doesn\'t accept a JSON body')

    def _validate_response(self, res, validate):
        if validate is True and res is not None:
            response_schema = self._method_specs.get('response')
            if response_schema is not None:
                if '$ref' in response_schema:
                    response_schema = self._schemas[response_schema['$ref']]
            else:
                raise ValidationError('Response body should\'ve been validated, but wasn\'t because a schema body wasn\'nt found')
            self._validate(res, response_schema)
        return res

    def __contains__(self, item):
        return item in self.parameters

    def __str__(self):
        return self['id'] + ' method @ ' + self._base_url

    def __repr__(self):
        return self.__str__()

    def __len__(self):
        return len(self.required_parameters) if self.required_parameters else 0


class Resource:
    def __init__(self, name, resource_specs, global_parameters, schemas, root_url, service_path, batch_path, validate):
        self.name = name
        self._resource_specs = resource_specs
        self._global_parameters = global_parameters
        self._schemas = schemas
        self._root_url = root_url
        self._service_path = service_path
        self._batch_path = batch_path
        self._validate = validate  

    @property
    def methods_available(self) -> [str, str]:
        '''
        Returns the names of the methods that this resource provides
        '''
        return [k for k,v in self['methods'].items()] if self['methods'] else []

    @property
    def resources_available(self) -> [str, str]:
        '''
        Returns the names of the nested resources in this resource
        '''
        return [k for k,v in self['resources'].items()] if self['resources'] else []

    def __str__(self):
        return self.name + ' resource @ ' + self._root_url + self._service_path

    def __repr__(self):
        return self.__str__()

    def __call__(self):
        raise TypeError('Only methods are callables, not resources.'
                        ' e.g. api.resource.list() NOT api.resource().list()')

    def __len__(self):
        return len(self.methods_available)

    def __contains__(self, item):
        return ((item in self.methods_available) or (item in self.resources_available))

    def __getitem__(self, k):
        return self._resource_specs.get(k)

    def __getattr__(self, method_or_resource):
        '''
        Returns either a method or a nested resource

        Arguments:

            method_or_resource: Name of the method or resource desired.

        Returns:

            aiogoogle.resource.Resource, aiogoogle.resource.Methods: A Resource or a Method


        Note:
            
            This method will first check in nested resources then will check in methods.

        Raises:

            AttributeError:
        '''
        # 1. Search in nested resources
        if method_or_resource in self.resources_available:
            return Resource(
                name=method_or_resource,
                resource_specs=self['resources'][method_or_resource],
                global_parameters=self._global_parameters,
                schemas=self._schemas,
                root_url=self._root_url,
                service_path=self._service_path,
                batch_path=self._batch_path,
                validate=self._validate
            )
        # 2. Search in methods
        elif method_or_resource in self.methods_available:
            return Method(
                name=method_or_resource,
                method_specs=self['methods'][method_or_resource],
                global_parameters=self._global_parameters,
                schemas=self._schemas,
                root_url=self._root_url,
                service_path=self._service_path,
                batch_path=self._batch_path,
                validate=self._validate
            )
        else:
            raise AttributeError(f"""Resource/Method {method_or_resource} doesn't exist.
                                 Check: https://developers.google.com/ for more info. 
                                 \nAvailable resources are: 
                                 {self.resources_available}\n
                                 Available methods are {self.methods_available}""")

class GoogleAPI:
    '''
    Creetes a representation of Google API given a discovery document
    
    Arguments:

        discovery_document (dict): A discovery document

        validate (bool): Whther or not to validate user input again the schema defined in the discovery document
    '''
    def __init__(self, discovery_document, validate=True):
        self.discovery_document = self._add_extra_params(discovery_document)
        self._validate = validate

    def _add_extra_params(self, discovery_document):
        ''' Adds extra parameters that aren't mentioned in the discovery docuemnt '''
        extra_params = {param: STACK_QUERY_PARAMETER_DEFAULT_VALUE for param in STACK_QUERY_PARAMETERS}
        if discovery_document.get('parameters'):
            discovery_document['parameters'] = {**discovery_document['parameters'], **extra_params}
        else:
            discovery_document['parameters'] = extra_params
        return discovery_document

    @property
    def methods_available(self) -> [str, str]:
        '''
        Returns names of the methods provided by this resource
        '''
        return [k for k,v in self['methods'].items()] if self['methods'] else []

    @property
    def resources_available(self) -> [str, str]:
        '''
        Returns names of the resources in a given API if any
        '''
        return [k for k,v in self['resources'].items()] if self['resources'] else []

    def __getattr__(self, method_or_resource) -> Resource:
        '''
        Returns resources from an API

        Note:
            
            This method will first check in resources then will check in methods.

        Arguments:

            method_or_resource (str): name of the top level method or resource

        Example:

            ::

                >>> google_service = GoogleAPI(google_service_discovery_doc)
                >>> google_service.user

                user resource @ google_service.com/api/

        Returns:

            aiogoogle.resource.Resource, aiogoogle.resource.Method: A Resource or a Method

        Raises:

            AttributeError:
        '''
        if method_or_resource in self.resources_available:
            return Resource(
                name=method_or_resource,
                resource_specs=self['resources'][method_or_resource],
                global_parameters=self['parameters'],
                schemas=self['schemas'] or {},  # josnschema validator will fail if schemas isn't a dict
                root_url=self['rootUrl'],
                service_path=self['servicePath'],
                batch_path=self['batchPath'],
                validate=self._validate
            )
        elif method_or_resource in self.methods_available:
            return Method(
                name=method_or_resource,
                method_specs=self['methods'][method_or_resource],
                global_parameters=self['parameters'],
                schemas=self['schemas'] or {},  # josnschema validator will fail if schemas isn't a dict
                root_url=self['rootUrl'],
                service_path=self['servicePath'],
                batch_path=self['batchPath'],
                validate=self._validate
            )
        else:
            documentation_link = self.discovery_document.get('documentationLink') or 'https://developers.google.com/'
            raise AttributeError(f"""Resource/Method {method_or_resource} doesn't exist.
                                 Check: {documentation_link} for more info. 
                                 \nAvailable resources are: 
                                 {self.resources_available}\n
                                 Available methods are {self.methods_available}""")

    def __getitem__(self, k):
        '''
        Returns items from the discovery document

        Example:

            ::

                >>> google_service = GoogleAPI(google_books_discovery_doc)
                >>> google_service['name']

                'books'

                >>> google_service['id']

                'books:v1'

                >>> google_service['version']

                'v1'

                >>> google_service['documentationLink']

                'https://developers.google.com/books/docs/v1/getting_started'                
                
                >>> google_service['oauth2']['scopes']

                https://www.googleapis.com/auth/books: {
                    
                    description: "Manage your books"
                
                }

        Returns:

            dict: Discovery Document Item
        '''
        return self.discovery_document.get(k)

    def __contains__(self, item):
        return (item in self.resources_available) or (item in self.methods_available)
    
    def __repr__(self):
        labels = f'\nLabels:\n{self["labels"]}' if self['labels'] is not None else ''
        return self.discovery_document['name'] + '-' + self.discovery_document['version'] + ' API @ ' + self['rootUrl'] + self['servicePath'] + labels

    def __str__(self):
        return self.__repr__()

    def __len__(self):
        return len(self.resources_available) + len(self.methods_available)

    def __call__(self):
        raise TypeError('Only methods are callables, not resources.'
                        ' e.g. client.resources.user.list() NOT client.resources().user.list()')



DISCOVERY_SERVICE_V1_DISCOVERY_DOC = {'basePath': '/discovery/v1/',
 'baseUrl': 'https://www.googleapis.com/discovery/v1/',
 'batchPath': 'batch/discovery/v1',
 'description': 'Provides information about other Google APIs, such as what APIs are available, the resource, and method details for each API.',
 'discoveryVersion': 'v1',
 'documentationLink': 'https://developers.google.com/discovery/',
 'etag': '"J3WqvAcMk4eQjJXvfSI4Yr8VouA/kUro_lCKdlL_SgLaWY4jaDKtqsk"',
 'icons': {'x16': 'http://www.google.com/images/icons/feature/filing_cabinet_search-g16.png', 'x32': 'http://www.google.com/images/icons/feature/filing_cabinet_search-g32.png'},
 'id': 'discovery:v1',
 'kind': 'discovery#restDescription',
 'name': 'discovery',
 'ownerDomain': 'google.com',
 'ownerName': 'Google',
 'parameters': {'alt': {'default': 'json',
                        'description': 'Data format for the response.',
                        'enum': ['json'],
                        'enumDescriptions': ['Responses with Content-Type of application/json'],
                        'location': 'query',
                        'type': 'string'},
                'fields': {'description': 'Selector specifying which fields to include in a partial response.', 'location': 'query', 'type': 'string'},
                'key': {'description': 'API key. Your API key identifies your project and provides you with API access, quota, and reports. Required unless you provide an OAuth 2.0 token.',
                        'location': 'query',
                        'type': 'string'},
                'oauth_token': {'description': 'OAuth 2.0 token for the current user.', 'location': 'query', 'type': 'string'},
                'prettyPrint': {'default': 'true', 'description': 'Returns response with indentations and line breaks.', 'location': 'query', 'type': 'boolean'},
                'quotaUser': {'description': 'An opaque string that represents a user for quota purposes. Must not exceed 40 characters.', 'location': 'query', 'type': 'string'},
                'userIp': {'description': 'Deprecated. Please use quotaUser instead.', 'location': 'query', 'type': 'string'}},
 'protocol': 'rest',
 'resources': {'apis': {'methods': {'getRest': {'description': 'Retrieve the description of a particular version of an api.',
                                                'httpMethod': 'GET',
                                                'id': 'discovery.apis.getRest',
                                                'parameterOrder': ['api', 'version'],
                                                'parameters': {'api': {'description': 'The name of the API.', 'location': 'path', 'required': True, 'type': 'string'},
                                                               'version': {'description': 'The version of the API.', 'location': 'path', 'required': True, 'type': 'string'}},
                                                'path': 'apis/{api}/{version}/rest',
                                                'response': {'$ref': 'RestDescription'}},
                                    'list': {'description': 'Retrieve the list of APIs supported at this endpoint.',
                                             'httpMethod': 'GET',
                                             'id': 'discovery.apis.list',
                                             'parameters': {'name': {'description': 'Only include APIs with the given name.', 'location': 'query', 'type': 'string'},
                                                            'preferred': {'default': 'false', 'description': 'Return only the preferred version of an API.', 'location': 'query', 'type': 'boolean'}},
                                             'path': 'apis',
                                             'response': {'$ref': 'DirectoryList'}}}}},
 'revision': '20181107',
 'rootUrl': 'https://www.googleapis.com/',
 'schemas': {'DirectoryList': {'id': 'DirectoryList',
                               'properties': {'discoveryVersion': {'default': 'v1', 'description': 'Indicate the version of the Discovery API used to generate this doc.', 'type': 'string'},
                                              'items': {'description': 'The individual directory entries. One entry per api/version pair.',
                                                        'items': {'properties': {'description': {'description': 'The description of this API.', 'type': 'string'},
                                                                                 'discoveryLink': {'description': 'A link to the discovery document.', 'type': 'string'},
                                                                                 'discoveryRestUrl': {'description': 'The URL for the discovery REST document.', 'type': 'string'},
                                                                                 'documentationLink': {'description': 'A link to human readable documentation for the API.', 'type': 'string'},
                                                                                 'icons': {'description': 'Links to 16x16 and 32x32 icons representing the API.',
                                                                                           'properties': {'x16': {'description': 'The URL of the 16x16 icon.', 'type': 'string'},
                                                                                                          'x32': {'description': 'The URL of the 32x32 icon.', 'type': 'string'}},
                                                                                           'type': 'object'},
                                                                                 'id': {'description': 'The id of this API.', 'type': 'string'},
                                                                                 'kind': {'default': 'discovery#directoryItem', 'description': 'The kind for this response.', 'type': 'string'},
                                                                                 'labels': {'description': 'Labels for the status of this API, such as labs or deprecated.',
                                                                                            'items': {'type': 'string'},
                                                                                            'type': 'array'},
                                                                                 'name': {'description': 'The name of the API.', 'type': 'string'},
                                                                                 'preferred': {'description': 'True if this version is the preferred version to use.', 'type': 'boolean'},
                                                                                 'title': {'description': 'The title of this API.', 'type': 'string'},
                                                                                 'version': {'description': 'The version of the API.', 'type': 'string'}},
                                                                  'type': 'object'},
                                                        'type': 'array'},
                                              'kind': {'default': 'discovery#directoryList', 'description': 'The kind for this response.', 'type': 'string'}},
                               'type': 'object'},
             'JsonSchema': {'id': 'JsonSchema',
                            'properties': {'$ref': {'description': 'A reference to another schema. The value of this property is the "id" of another schema.', 'type': 'string'},
                                           'additionalProperties': {'$ref': 'JsonSchema',
                                                                    'description': 'If this is a schema for an object, this property is the schema for any additional properties with dynamic keys on '
                                                                                   'this object.'},
                                           'annotations': {'description': 'Additional information about this property.',
                                                           'properties': {'required': {'description': 'A list of methods for which this property is required on requests.',
                                                                                       'items': {'type': 'string'},
                                                                                       'type': 'array'}},
                                                           'type': 'object'},
                                           'default': {'description': 'The default value of this property (if one exists).', 'type': 'string'},
                                           'description': {'description': 'A description of this object.', 'type': 'string'},
                                           'enum': {'description': 'Values this parameter may take (if it is an enum).', 'items': {'type': 'string'}, 'type': 'array'},
                                           'enumDescriptions': {'description': 'The descriptions for the enums. Each position maps to the corresponding value in the "enum" array.',
                                                                'items': {'type': 'string'},
                                                                'type': 'array'},
                                           'format': {'description': 'An additional regular expression or key that helps constrain the value. For more details see: '
                                                                     'http://tools.ietf.org/html/draft-zyp-json-schema-03#section-5.23',
                                                      'type': 'string'},
                                           'id': {'description': 'Unique identifier for this schema.', 'type': 'string'},
                                           'items': {'$ref': 'JsonSchema', 'description': 'If this is a schema for an array, this property is the schema for each element in the array.'},
                                           'location': {'description': 'Whether this parameter goes in the query or the path for REST requests.', 'type': 'string'},
                                           'maximum': {'description': 'The maximum value of this parameter.', 'type': 'string'},
                                           'minimum': {'description': 'The minimum value of this parameter.', 'type': 'string'},
                                           'pattern': {'description': 'The regular expression this parameter must conform to. Uses Java 6 regex format: '
                                                                      'http://docs.oracle.com/javase/6/docs/api/java/util/regex/Pattern.html',
                                                       'type': 'string'},
                                           'properties': {'additionalProperties': {'$ref': 'JsonSchema',
                                                                                   'description': 'A single property of this object. The value is itself a JSON Schema object describing this '
                                                                                                  'property.'},
                                                          'description': 'If this is a schema for an object, list the schema for each property of this object.',
                                                          'type': 'object'},
                                           'readOnly': {'description': 'The value is read-only, generated by the service. The value cannot be modified by the client. If the value is included in a '
                                                                       'POST, PUT, or PATCH request, it is ignored by the service.',
                                                        'type': 'boolean'},
                                           'repeated': {'description': 'Whether this parameter may appear multiple times.', 'type': 'boolean'},
                                           'required': {'description': 'Whether the parameter is required.', 'type': 'boolean'},
                                           'type': {'description': 'The value type for this schema. A list of values can be found here: '
                                                                   'http://tools.ietf.org/html/draft-zyp-json-schema-03#section-5.1',
                                                    'type': 'string'},
                                           'variant': {'description': 'In a variant data type, the value of one property is used to determine how to interpret the entire entity. Its value must exist '
                                                                      'in a map of descriminant values to schema names.',
                                                       'properties': {'discriminant': {'description': 'The name of the type discriminant property.', 'type': 'string'},
                                                                      'map': {'description': 'The map of discriminant value to schema to use for parsing..',
                                                                              'items': {'properties': {'$ref': {'type': 'string'}, 'type_value': {'type': 'string'}}, 'type': 'object'},
                                                                              'type': 'array'}},
                                                       'type': 'object'}},
                            'type': 'object'},
             'RestDescription': {'id': 'RestDescription',
                                 'properties': {'auth': {'description': 'Authentication information.',
                                                         'properties': {'oauth2': {'description': 'OAuth 2.0 authentication information.',
                                                                                   'properties': {'scopes': {'additionalProperties': {'description': 'The scope value.',
                                                                                                                                      'properties': {'description': {'description': 'Description of '
                                                                                                                                                                                    'scope.',
                                                                                                                                                                     'type': 'string'}},
                                                                                                                                      'type': 'object'},
                                                                                                             'description': 'Available OAuth 2.0 scopes.',
                                                                                                             'type': 'object'}},
                                                                                   'type': 'object'}},
                                                         'type': 'object'},
                                                'basePath': {'description': '[DEPRECATED] The base path for REST requests.', 'type': 'string'},
                                                'baseUrl': {'description': '[DEPRECATED] The base URL for REST requests.', 'type': 'string'},
                                                'batchPath': {'description': 'The path for REST batch requests.', 'type': 'string'},
                                                'canonicalName': {'description': 'Indicates how the API name should be capitalized and split into various parts. Useful for generating pretty class '
                                                                                 'names.',
                                                                  'type': 'string'},
                                                'description': {'description': 'The description of this API.', 'type': 'string'},
                                                'discoveryVersion': {'default': 'v1', 'description': 'Indicate the version of the Discovery API used to generate this doc.', 'type': 'string'},
                                                'documentationLink': {'description': 'A link to human readable documentation for the API.', 'type': 'string'},
                                                'etag': {'description': 'The ETag for this response.', 'readOnly': True, 'type': 'string'},
                                                'exponentialBackoffDefault': {'description': 'Enable exponential backoff for suitable methods in the generated clients.', 'type': 'boolean'},
                                                'features': {'description': 'A list of supported features for this API.', 'items': {'type': 'string'}, 'type': 'array'},
                                                'icons': {'description': 'Links to 16x16 and 32x32 icons representing the API.',
                                                          'properties': {'x16': {'description': 'The URL of the 16x16 icon.', 'type': 'string'},
                                                                         'x32': {'description': 'The URL of the 32x32 icon.', 'type': 'string'}},
                                                          'type': 'object'},
                                                'id': {'description': 'The ID of this API.', 'type': 'string'},
                                                'kind': {'default': 'discovery#restDescription', 'description': 'The kind for this response.', 'type': 'string'},
                                                'labels': {'description': 'Labels for the status of this API, such as labs or deprecated.', 'items': {'type': 'string'}, 'type': 'array'},
                                                'methods': {'additionalProperties': {'$ref': 'RestMethod', 'description': 'An individual method description.'},
                                                            'description': 'API-level methods for this API.',
                                                            'type': 'object'},
                                                'name': {'description': 'The name of this API.', 'type': 'string'},
                                                'ownerDomain': {'description': 'The domain of the owner of this API. Together with the ownerName and a packagePath values, this can be used to '
                                                                               'generate a library for this API which would have a unique fully qualified name.',
                                                                'type': 'string'},
                                                'ownerName': {'description': 'The name of the owner of this API. See ownerDomain.', 'type': 'string'},
                                                'packagePath': {'description': 'The package of the owner of this API. See ownerDomain.', 'type': 'string'},
                                                'parameters': {'additionalProperties': {'$ref': 'JsonSchema', 'description': 'Description of a single parameter.'},
                                                               'description': 'Common parameters that apply across all apis.',
                                                               'type': 'object'},
                                                'protocol': {'default': 'rest', 'description': 'The protocol described by this document.', 'type': 'string'},
                                                'resources': {'additionalProperties': {'$ref': 'RestResource',
                                                                                       'description': 'An individual resource description. Contains methods and sub-resources related to this '
                                                                                                      'resource.'},
                                                              'description': 'The resources in this API.',
                                                              'type': 'object'},
                                                'revision': {'description': 'The version of this API.', 'type': 'string'},
                                                'rootUrl': {'description': 'The root URL under which all API services live.', 'type': 'string'},
                                                'schemas': {'additionalProperties': {'$ref': 'JsonSchema', 'description': 'An individual schema description.'},
                                                            'description': 'The schemas for this API.',
                                                            'type': 'object'},
                                                'servicePath': {'description': 'The base path for all REST requests.', 'type': 'string'},
                                                'title': {'description': 'The title of this API.', 'type': 'string'},
                                                'version': {'description': 'The version of this API.', 'type': 'string'},
                                                'version_module': {'type': 'boolean'}},
                                 'type': 'object'},
             'RestMethod': {'id': 'RestMethod',
                            'properties': {'description': {'description': 'Description of this method.', 'type': 'string'},
                                           'etagRequired': {'description': 'Whether this method requires an ETag to be specified. The ETag is sent as an HTTP If-Match or If-None-Match header.',
                                                            'type': 'boolean'},
                                           'httpMethod': {'description': 'HTTP method used by this method.', 'type': 'string'},
                                           'id': {'description': 'A unique ID for this method. This property can be used to match methods between different versions of Discovery.', 'type': 'string'},
                                           'mediaUpload': {'description': 'Media upload parameters.',
                                                           'properties': {'accept': {'description': 'MIME Media Ranges for acceptable media uploads to this method.',
                                                                                     'items': {'type': 'string'},
                                                                                     'type': 'array'},
                                                                          'maxSize': {'description': 'Maximum size of a media upload, such as "1MB", "2GB" or "3TB".', 'type': 'string'},
                                                                          'protocols': {'description': 'Supported upload protocols.',
                                                                                        'properties': {'resumable': {'description': 'Supports the Resumable Media Upload protocol.',
                                                                                                                     'properties': {'multipart': {'default': 'true',
                                                                                                                                                  'description': 'True if this endpoint supports '
                                                                                                                                                                 'uploading multipart media.',
                                                                                                                                                  'type': 'boolean'},
                                                                                                                                    'path': {'description': 'The URI path to be used for upload. '
                                                                                                                                                            'Should be used in conjunction with the '
                                                                                                                                                            'basePath property at the api-level.',
                                                                                                                                             'type': 'string'}},
                                                                                                                     'type': 'object'},
                                                                                                       'simple': {'description': 'Supports uploading as a single HTTP request.',
                                                                                                                  'properties': {'multipart': {'default': 'true',
                                                                                                                                               'description': 'True if this endpoint supports upload '
                                                                                                                                                              'multipart media.',
                                                                                                                                               'type': 'boolean'},
                                                                                                                                 'path': {'description': 'The URI path to be used for upload. Should '
                                                                                                                                                         'be used in conjunction with the basePath '
                                                                                                                                                         'property at the api-level.',
                                                                                                                                          'type': 'string'}},
                                                                                                                  'type': 'object'}},
                                                                                        'type': 'object'}},
                                                           'type': 'object'},
                                           'parameterOrder': {'description': 'Ordered list of required parameters, serves as a hint to clients on how to structure their method signatures. The array '
                                                                             'is ordered such that the "most-significant" parameter appears first.',
                                                              'items': {'type': 'string'},
                                                              'type': 'array'},
                                           'parameters': {'additionalProperties': {'$ref': 'JsonSchema', 'description': 'Details for a single parameter in this method.'},
                                                          'description': 'Details for all parameters in this method.',
                                                          'type': 'object'},
                                           'path': {'description': 'The URI path of this REST method. Should be used in conjunction with the basePath property at the api-level.', 'type': 'string'},
                                           'request': {'description': 'The schema for the request.',
                                                       'properties': {'$ref': {'description': 'Schema ID for the request schema.', 'type': 'string'},
                                                                      'parameterName': {'description': 'parameter name.', 'type': 'string'}},
                                                       'type': 'object'},
                                           'response': {'description': 'The schema for the response.',
                                                        'properties': {'$ref': {'description': 'Schema ID for the response schema.', 'type': 'string'}},
                                                        'type': 'object'},
                                           'scopes': {'description': 'OAuth 2.0 scopes applicable to this method.', 'items': {'type': 'string'}, 'type': 'array'},
                                           'supportsMediaDownload': {'description': 'Whether this method supports media downloads.', 'type': 'boolean'},
                                           'supportsMediaUpload': {'description': 'Whether this method supports media uploads.', 'type': 'boolean'},
                                           'supportsSubscription': {'description': 'Whether this method supports subscriptions.', 'type': 'boolean'},
                                           'useMediaDownloadService': {'description': 'Indicates that downloads from this method should use the download service URL (i.e. "/download"). Only applies '
                                                                                      'if the method supports media download.',
                                                                       'type': 'boolean'}},
                            'type': 'object'},
             'RestResource': {'id': 'RestResource',
                              'properties': {'methods': {'additionalProperties': {'$ref': 'RestMethod', 'description': 'Description for any methods on this resource.'},
                                                         'description': 'Methods on this resource.',
                                                         'type': 'object'},
                                             'resources': {'additionalProperties': {'$ref': 'RestResource', 'description': 'Description for any sub-resources on this resource.'},
                                                           'description': 'Sub-resources on this resource.',
                                                           'type': 'object'}},
                              'type': 'object'}},
 'servicePath': 'discovery/v1/',
 'title': 'APIs Discovery Service',
 'version': 'v1'}


class AiogoogleError(Exception):
    pass

class ValidationError(AiogoogleError):
    '''
    Raised when the validate flag is set true and validation error occurs
    '''
    pass

class HTTPError(AiogoogleError):
    def __init__(self, msg, req=None, res=None):
        self.req = req
        self.res = res
        super().__init__(msg)

class AuthError(HTTPError):
    pass


import setuptools

import os
from os import listdir
from os.path import isfile, join
mypath = os.path.dirname(os.path.abspath(__file__))
print([f for f in listdir(mypath) if isfile(join(mypath, f))])


with open('requirements.txt', "r") as f:
    requirements = f.read().splitlines()

with open('test_requirements.txt', "r") as f:
    test_requirements = f.read().splitlines()

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(

    name='aiogoogle',
    version="0.0.1b1",
    author='Omar Ryhan',
    author_email='omarryhan@gmail.com',
    license='GNU',
    description="Aiogoogle is an async-framework agnostic async library for Google's Discovery Service",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    tests_require=test_requirements,
    url='https://github.com/omarryhan/aiogoogle',
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Operating System :: OS Independent",
    ],
    extras_require={
        'curio_asks': ["asks", "curio"],
        'trio_asks': ["asks", "trio"]
    },
)


from aiogoogle.auth.managers import ApiKeyManager
from aiogoogle.models import Request
from aiogoogle.resource import GoogleAPI
import pytest


@pytest.fixture
def manager():
    return ApiKeyManager()

@pytest.fixture
def key1():
    return 'asdasdasidajosnoqwndoasfnwuiefn'

@pytest.fixture
def key2():
    return 'asjdaosdniaurn3uh2873he9has9dhASDQ(Dasd'


def test_google_api_is_functional(manager, create_api, key1):
    api = create_api('urlshortener', 'v1')
    request_with_key = api.url.get(key=key1, shortUrl='asdasd', validate=False)
    assert key1 in request_with_key.url

#--- tests begin here

def test_doesnt_add_key_when_existing(manager, create_api, key1):
    api = create_api('urlshortener', 'v1')
    request_with_key = api.url.get(key=key1, shortUrl='asdasd', validate=False)
    new_request = manager.authorize(request_with_key, key1)
    assert request_with_key.url == new_request.url

def test_adds_key_to_url_with_query(manager, create_api, key1):
    api = create_api('urlshortener', 'v1')
    request_with_key = api.url.get(shortUrl='asdasd', key=key1)  # Must be kept in this order for this test to pass
    request_without_key = api.url.get(shortUrl='asdasd')
    request_without_key_authorized = manager.authorize(request_without_key, key1)
    assert request_without_key_authorized.url == request_with_key.url

def test_adds_key_to_url_without_query(manager, key1):
    url = 'https://example.com/api'
    req = Request(url=url)
    request_with_key = manager.authorize(req, key1)
    assert request_with_key.url == url + '?key=' + key1

def test_adds_key_to_url_without_query_and_with_slash(manager, key1):
    url = 'https://example.com/api/'
    req = Request(url=url)
    request_with_key = manager.authorize(req, key1)
    assert request_with_key.url == url[:-1] + '?key=' + key1
    
    


__all__ = ['AbstractSession']


from abc import ABC, abstractproperty, abstractmethod, abstractclassmethod, abstractstaticmethod
import inspect


class AbstractSession(ABC):
    '''
    Should represent an HTTP session that has an asynchronous context manager and a send method

    Argumetnts:

        Should return an instance without needing to pass it any args or kwargs
    '''

    def __new__(cls, *args, **kwargs):
        # Get all coros of this the abstract class
        parent_abstract_coros = inspect.getmembers(AbstractSession, predicate=inspect.iscoroutinefunction)

        # Ensure all relevant child methods are implemented as coros
        for coro in parent_abstract_coros:
            coro_name = coro[0]
            child_method = getattr(cls, coro_name)
            if not inspect.iscoroutinefunction(child_method):
                raise RuntimeError(f'{child_method} must be a coroutine')

        # Resume with normal behavior of a Python constructor
        return super(AbstractSession, cls).__new__(cls)

    @abstractmethod
    async def send(self, *requests, timeout=None,  full_res=False, session_factory=None):
        '''
        Takes requests, sends them, returns contents of responses or full http responses.

        Note:

            Given more than one request this method should return responses as a list. However if only given one, it will return a single response object
        
        Arguments:
            
            requests (aiogoogle.models.Request):

                Request objects from aiogoogle.models
            
            timeout (int):

                Total timeout for *requests

            full_res (bool):

                Return full HTTP response with headers, status code

            raise_for_status (bool):

                * Whether or not to raise an HTTP error if status code is >= 400

                * Default: True

            session_factory (aiogoogle.sessions.abs.AbstractSession):

                * A callable implementation of aiogoogle's session interface

                * Session factory that creates a session that will handle pagination, resumable uploads etc.

                * Defaults to ``self.__class__`` 

        Returns:

            aiogoogle.models.Response

        Raises:

            aiogoogle.excs.HTTPError

        '''
        raise NotImplementedError


import os, sys
import pytest

args = [
    '-v',
    '-s',
    '--cov',
    'aiogoogle/',
    #'--tb=long',
    #'--fulltrace',
    #'--maxfail=5',
    #'-p no:warnings',
    #'--disable-warnings',
]

def main():

    if os.getenv('TEST_UNITS') == 'true':
        args.append(
            'tests/test_units/',
        )
    if os.getenv('TEST_INTEGRATION_OFFLINE') == 'true':
        args.append(
            'tests/test_integ_offline/',
        )
    if os.getenv('TEST_INTEGRATION_ONLINE') == 'true':
        args.append(
            'tests/test_integ_online/',
        )
    if os.getenv('TEST_INTEGRATION_WITH_KEYS') == 'true':
        args.append(
            'tests/test_with_keys/'
        )
    sys.exit(pytest.main(args))

if __name__ == '__main__':
    main()